# -*- coding: utf-8 -*-

"""
This file contains a basic script class to run with qudi module dependencies as well as various
helper classes to run and manage these scripts.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['import_module_script', 'ModuleScript', 'ModuleScriptsDictTableModel',
           'ModuleScriptInterrupted']

import importlib
import copy
import inspect
from abc import abstractmethod
from PySide2 import QtCore
from logging import Logger
from typing import Mapping, Any, Type, Optional, Union, Dict

from qudi.core.meta import QudiObjectMeta
from qudi.core.module import Base
from qudi.core.logger import get_logger
from qudi.util.models import DictTableModel
from qudi.util.mutex import Mutex


class ModuleScriptInterrupted(Exception):
    """Custom exception class to indicate that a ModuleScript execution has been interrupted.
    """
    pass


class ModuleScript(QtCore.QObject, metaclass=QudiObjectMeta):
    """
    The only part that can be interrupted is the _run() method.
    The implementations must occasionally call _check_interrupt() to raise an exception at that
    point if an interrupt is requested.
    """
    # Declare all module connectors used in this script here

    sigFinished = QtCore.Signal()

    # FIXME: This __new__ implementation has the sole purpose to circumvent a known PySide2(6) bug.
    #  See https://bugreports.qt.io/browse/PYSIDE-1434 for more details.
    def __new__(cls, *args, **kwargs):
        abstract = getattr(cls, '__abstractmethods__', frozenset())
        if abstract:
            raise TypeError(f'Can\'t instantiate abstract class "{cls.__name__}" '
                            f'with abstract methods {set(abstract)}')
        return super().__new__(cls, *args, **kwargs)

    def __init__(self):
        # ModuleScript QObjects must not have a parent in order to be used as threaded workers
        super().__init__()

        # Create a copy of the _meta class dict and attach it to this instance
        self._meta = copy.deepcopy(self._meta)
        # set instance attributes according to connector meta objects
        for attr_name, conn in self._meta['connectors'].items():
            setattr(self, attr_name, conn)

        self._thread_lock = Mutex()
        self.__logger = get_logger(f'{self.__module__}.{self.__class__.__name__}')

        # script arguments and result cache
        self.args = tuple()
        self.kwargs = dict()
        self.result = None

        # Status flags
        self._interrupted = True
        self._success = False
        self._running = False

    @property
    def interrupted(self):
        with self._thread_lock:
            return self._interrupted

    @property
    def log(self) -> Logger:
        """Returns a logger object.
        DO NOT OVERRIDE IN SUBCLASS!

        Returns
        -------
        Logger
            Logger object for this script class.
        """
        return self.__logger

    @property
    def running(self) -> bool:
        with self._thread_lock:
            return self._running

    @property
    def success(self) -> bool:
        with self._thread_lock:
            return self._success

    @property
    def connected_modules(self) -> Mapping[str, Union[str, None]]:
        """Mapping of Connector names (keys) to connected module target names (values).
        Unconnected Connectors are indicated by None target.
        """
        return {conn.name: None if conn() is None else conn().module_name for conn in
                self._meta['connectors']}

    @classmethod
    def call_parameters(cls) -> Dict[str, inspect.Parameter]:
        """Call parameters of the _run method implementation.

        Override in subclass if you want anything else than this default implementation.
        Make sure custom implementations of this property are compatible with _run!
        """
        parameters = dict(inspect.signature(cls._run).parameters)
        # Remove first parameter if it is a bound instance method
        if not isinstance(inspect.getattr_static(cls, '_run'), (classmethod, staticmethod)):
            try:
                del parameters[next(iter(parameters))]
            except StopIteration:
                pass
        return parameters

    @classmethod
    def result_annotation(cls) -> Union[Any, inspect.Signature.empty]:
        """Return type annotation for the _run method implementation.
        Will return inspect.Signature.empty if _run return value is not annotated.
        """
        return inspect.signature(cls._run).return_annotation

    def interrupt(self):
        with self._thread_lock:
            self._interrupted = True

    def __call__(self, *args, **kwargs) -> Any:
        """Convenience magic method to run this script like a function.
        DO NOT OVERRIDE IN SUBCLASS!

        Arguments are passed directly to _run() method.

        Returns
        -------
        object
            Result of the script method.
        """
        self.args = args
        self.kwargs = kwargs
        self.run()
        return self.result

    @QtCore.Slot()
    def run(self) -> None:
        """Check run prerequisites and execute _run method with pre-cached arguments.
        DO NOT OVERRIDE IN SUBCLASS!
        """
        self.result = None
        with self._thread_lock:
            self._interrupted = False
            self._success = False
            self._running = True
        self.log.debug(f'Running main method with\n\targs: {self.args}\n\tkwargs: {self.kwargs}.')
        # Emit finished signal even if script execution fails. Check success flag.
        try:
            self.result = self._run(*self.args, **self.kwargs)
            with self._thread_lock:
                self._success = True
        except ModuleScriptInterrupted:
            self.log.info(f'Main run method interrupted')
        except:
            self.log.exception('Exception during main run method:')
            raise
        finally:
            with self._thread_lock:
                self._running = False
                self.sigFinished.emit()

    def connect_modules(self, connector_targets: Mapping[str, Base]) -> None:
        """Connects given modules (values) to their respective Connector (keys).

        DO NOT CALL THIS METHOD UNLESS YOU KNOW WHAT YOU ARE DOING!
        """
        # Sanity checks
        conn_names = set(conn.name for conn in self._meta['connectors'].values())
        mandatory_conn = set(
            conn.name for conn in self._meta['connectors'].values() if not conn.optional
        )
        configured_conn = set(connector_targets)
        if not configured_conn.issubset(conn_names):
            raise KeyError(f'Mismatch of connectors in configuration {configured_conn} and '
                           f'Connector meta objects {conn_names}.')
        if not mandatory_conn.issubset(configured_conn):
            raise ValueError(f'Not all mandatory connectors are specified.\n'
                             f'Mandatory connectors are: {mandatory_conn}')

        # Iterate through module connectors and connect them if possible
        for conn in self._meta['connectors'].values():
            target = connector_targets.get(conn.name, None)
            if target is None:
                continue
            if conn.is_connected:
                raise RuntimeError(f'Connector "{conn.name}" already connected.\n'
                                   f'Call "disconnect_modules()" before trying to reconnect.')
            conn.connect(target)

    def disconnect_modules(self) -> None:
        """Disconnects all Connector instances for this object.

        DO NOT CALL THIS METHOD UNLESS YOU KNOW WHAT YOU ARE DOING!
        """
        for conn in self._meta['connectors'].values():
            conn.disconnect()

    def _check_interrupt(self) -> None:
        """ Implementations of _run should occasionally call this method in order to break
        execution early if another thread has interrupted this script in the meantime.
        """
        if self.interrupted:
            raise ModuleScriptInterrupted

    @abstractmethod
    def _run(self, *args, **kwargs) -> Any:
        """The actual script to be run. Implement only this method in a subclass.
        """
        raise NotImplementedError(f'No _run() method implemented for "{self.__class__.__name__}".')


def import_module_script(module: str, cls: str,
                         reload: Optional[bool] = True) -> Type[ModuleScript]:
    """Helper function to import ModuleScript sub-classes by name from a given module.
    Reloads the module to import from by default.
    """
    mod = importlib.import_module(module)
    if reload:
        importlib.reload(mod)
    script = getattr(mod, cls)
    if not issubclass(script, ModuleScript):
        raise TypeError(f'Module script to import must be a subclass of {__name__}.ModuleScript')
    return script


class ModuleScriptsDictTableModel(DictTableModel):
    """Qt compatible table model holding all configured and available ModuleScript subclasses.
    """
    def __init__(self, scripts_config: Optional[Mapping[str, dict]] = None):
        super().__init__(headers=['Name', 'Class'])
        if scripts_config is None:
            scripts_config = dict()
        for name, config in scripts_config.items():
            self.register_script(name, config)

    def register_script(self, name: str, config: dict) -> None:
        if name in self:
            raise KeyError(f'Multiple module script with name "{name}" configured.')
        module, cls = config['module.Class'].rsplit('.', 1)
        self[name] = import_module_script(module, cls, reload=False)



# -*- coding: utf-8 -*-

"""
ToDo

Copyright (c) 2022, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from PySide2 import QtCore, QtWidgets, QtGui
from typing import Optional, Mapping, Any, Dict, Tuple, Union, List, Sequence
import pyqtgraph as pg

from qudi.util.widgets.scientific_spinbox import ScienDSpinBox
from qudi.util.widgets.separator_lines import VerticalLine
from qudi.util.widgets.plotting.axis import label_nudged_plot_widget
from qudi.util.widgets.plotting.plot_widget import RubberbandZoomSelectionPlotWidget
from qudi.util.units import ScaledFloat


PlotWidget = label_nudged_plot_widget(RubberbandZoomSelectionPlotWidget)


class PlotEditorWidget(QtWidgets.QWidget):
    """
    """

    sigAutoRangeClicked = QtCore.Signal(bool, bool)  # x- and/or y-axis
    sigLabelsChanged = QtCore.Signal(object, object)
    sigUnitsChanged = QtCore.Signal(object, object)
    sigLimitsChanged = QtCore.Signal(object, object)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)

        layout = QtWidgets.QGridLayout()
        self.setLayout(layout)

        # Generate labels
        x_label = QtWidgets.QLabel('Horizontal Axis:')
        x_label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        y_label = QtWidgets.QLabel('Vertical Axis:')
        y_label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        label_label = QtWidgets.QLabel('Label')
        label_label.setAlignment(QtCore.Qt.AlignCenter)
        unit_label = QtWidgets.QLabel('Units')
        unit_label.setAlignment(QtCore.Qt.AlignCenter)
        range_label = QtWidgets.QLabel('Range')
        range_label.setAlignment(QtCore.Qt.AlignCenter)
        # Generate editors
        self.x_label_lineEdit = QtWidgets.QLineEdit()
        self.x_label_lineEdit.setMinimumWidth(50)
        self.x_unit_lineEdit = QtWidgets.QLineEdit()
        self.x_unit_lineEdit.setMinimumWidth(50)
        self.x_lower_limit_spinBox = ScienDSpinBox()
        self.x_lower_limit_spinBox.setMinimumWidth(70)
        self.x_upper_limit_spinBox = ScienDSpinBox()
        self.x_upper_limit_spinBox.setMinimumWidth(70)
        self.x_auto_button = QtWidgets.QPushButton('Auto Range')
        self.y_label_lineEdit = QtWidgets.QLineEdit()
        self.y_label_lineEdit.setMinimumWidth(50)
        self.y_unit_lineEdit = QtWidgets.QLineEdit()
        self.y_unit_lineEdit.setMinimumWidth(50)
        self.y_lower_limit_spinBox = ScienDSpinBox()
        self.y_lower_limit_spinBox.setMinimumWidth(70)
        self.y_upper_limit_spinBox = ScienDSpinBox()
        self.y_upper_limit_spinBox.setMinimumWidth(70)
        self.y_auto_button = QtWidgets.QPushButton('Auto Range')

        row = 0
        layout.addWidget(label_label, row, 1)
        layout.addWidget(unit_label, row, 2)
        layout.addWidget(range_label, row, 4, 1, 2)
        row += 1
        layout.addWidget(x_label, row, 0)
        layout.addWidget(self.x_label_lineEdit, row, 1)
        layout.addWidget(self.x_unit_lineEdit, row, 2)
        layout.addWidget(self.x_lower_limit_spinBox, row, 4)
        layout.addWidget(self.x_upper_limit_spinBox, row, 5)
        layout.addWidget(self.x_auto_button, row, 6)
        row += 1
        layout.addWidget(y_label, row, 0)
        layout.addWidget(self.y_label_lineEdit, row, 1)
        layout.addWidget(self.y_unit_lineEdit, row, 2)
        layout.addWidget(self.y_lower_limit_spinBox, row, 4)
        layout.addWidget(self.y_upper_limit_spinBox, row, 5)
        layout.addWidget(self.y_auto_button, row, 6)
        row += 1
        layout.addWidget(VerticalLine(), 0, 3, row, 1)

        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(2, 1)
        layout.setColumnStretch(4, 3)
        layout.setColumnStretch(5, 3)

        self.x_label_lineEdit.editingFinished.connect(self.__x_label_changed)
        self.y_label_lineEdit.editingFinished.connect(self.__y_label_changed)
        self.x_unit_lineEdit.editingFinished.connect(self.__x_unit_changed)
        self.y_unit_lineEdit.editingFinished.connect(self.__y_unit_changed)
        self.x_lower_limit_spinBox.editingFinished.connect(self.__x_limits_changed)
        self.x_upper_limit_spinBox.editingFinished.connect(self.__x_limits_changed)
        self.y_lower_limit_spinBox.editingFinished.connect(self.__y_limits_changed)
        self.y_upper_limit_spinBox.editingFinished.connect(self.__y_limits_changed)
        self.x_auto_button.clicked.connect(
            lambda: self.sigAutoRangeClicked.emit(True, False)
        )
        self.y_auto_button.clicked.connect(
            lambda: self.sigAutoRangeClicked.emit(False, True)
        )

        self.set_limits((-0.5, 0.5), (-0.5, 0.5))
        self.set_units('arb.u.', 'arb.u.')
        self.set_labels('X', 'Y')

    @property
    def labels(self) -> Tuple[str, str]:
        return self.x_label_lineEdit.text(), self.y_label_lineEdit.text()

    @property
    def units(self) -> Tuple[str, str]:
        return self.x_unit_lineEdit.text(), self.y_unit_lineEdit.text()

    @property
    def limits(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        x_min, x_max = sorted([self.x_lower_limit_spinBox.value(),
                               self.x_upper_limit_spinBox.value()])
        y_min, y_max = sorted([self.y_lower_limit_spinBox.value(),
                               self.y_upper_limit_spinBox.value()])
        return (x_min, x_max), (y_min, y_max)

    def set_labels(self, x: Optional[str] = None, y: Optional[str] = None) -> None:
        if x is not None:
            self.x_label_lineEdit.setText(x)
        if y is not None:
            self.y_label_lineEdit.setText(y)

    def set_units(self, x: Optional[str] = None, y: Optional[str] = None) -> None:
        if x is not None:
            self.x_unit_lineEdit.setText(x)
        if y is not None:
            self.y_unit_lineEdit.setText(y)

    def set_limits(self,
                   x: Optional[Tuple[float, float]] = None,
                   y: Optional[Tuple[float, float]] = None
                   ) -> None:
        if x is not None:
            lower, upper = sorted(x)
            self.x_lower_limit_spinBox.setValue(lower)
            self.x_upper_limit_spinBox.setValue(upper)
        if y is not None:
            lower, upper = sorted(y)
            self.y_lower_limit_spinBox.setValue(lower)
            self.y_upper_limit_spinBox.setValue(upper)

    def __x_limits_changed(self) -> None:
        lower = self.x_lower_limit_spinBox.value()
        upper = self.x_upper_limit_spinBox.value()
        if upper < lower:
            lower, upper = upper, lower
            self.x_lower_limit_spinBox.setValue(lower)
            self.x_upper_limit_spinBox.setValue(upper)
            self.__swap_limits_focus()
        self.sigLimitsChanged.emit((lower, upper), None)

    def __y_limits_changed(self) -> None:
        lower = self.y_lower_limit_spinBox.value()
        upper = self.y_upper_limit_spinBox.value()
        if upper < lower:
            lower, upper = upper, lower
            self.y_lower_limit_spinBox.setValue(lower)
            self.y_upper_limit_spinBox.setValue(upper)
            self.__swap_limits_focus()
        self.sigLimitsChanged.emit(None, self.limits[1])

    def __x_label_changed(self) -> None:
        self.sigLabelsChanged.emit(self.labels[0], None)

    def __y_label_changed(self) -> None:
        self.sigLabelsChanged.emit(None, self.labels[1])

    def __x_unit_changed(self) -> None:
        self.sigUnitsChanged.emit(self.units[0], None)

    def __y_unit_changed(self) -> None:
        self.sigUnitsChanged.emit(None, self.units[1])

    def __swap_limits_focus(self) -> None:
        if self.x_lower_limit_spinBox.hasFocus():
            self.x_upper_limit_spinBox.setFocus()
        elif self.x_upper_limit_spinBox.hasFocus():
            self.x_lower_limit_spinBox.setFocus()
        elif self.y_lower_limit_spinBox.hasFocus():
            self.y_upper_limit_spinBox.setFocus()
        elif self.y_upper_limit_spinBox.hasFocus():
            self.y_lower_limit_spinBox.setFocus()


class PlotLegendIconWidget(QtWidgets.QWidget):
    def __init__(self, item, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)

        self.setMouseTracking(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.setFixedSize(20, 20)

        self._item = item

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        p = QtGui.QPainter(self)

        opts = self._item.opts
        if opts.get('antialias'):
            p.setRenderHint(p.RenderHint.Antialiasing)

        if not isinstance(self._item, pg.ScatterPlotItem):
            p.setPen(pg.mkPen(opts['pen']))
            p.drawLine(0, 11, 20, 11)

            if (opts.get('fillLevel', None) is not None and
                    opts.get('fillBrush', None) is not None):
                p.setBrush(pg.mkBrush(opts['fillBrush']))
                p.setPen(pg.mkPen(opts['pen']))
                p.drawPolygon(QtGui.QPolygonF(
                    [QtCore.QPointF(2, 18), QtCore.QPointF(18, 2),
                     QtCore.QPointF(18, 18)]))

        symbol = opts.get('symbol', None)
        if symbol is not None:
            if isinstance(self._item, pg.PlotDataItem):
                opts = self._item.scatter.opts
            p.translate(10, 10)
            pg.graphicsItems.ScatterPlotItem.drawSymbol(p, symbol, opts['size'], pg.mkPen(opts['pen']), pg.mkBrush(opts['brush']))

        if isinstance(self._item, pg.BarGraphItem):
            p.setBrush(pg.mkBrush(opts['brush']))
            p.drawRect(QtCore.QRectF(2, 2, 18, 18))


class PlotSelectorWidget(QtWidgets.QWidget):
    """
    """
    sigSelectionChanged = QtCore.Signal(dict)  # selection

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)

        self._stretch = QtWidgets.QSpacerItem(
            0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
        )
        self._selector_layout = QtWidgets.QGridLayout()
        self._selector_layout.addItem(self._stretch, 0, 0, 1, 2)
        self._selector_layout.setColumnStretch(0, 1)
        self.setLayout(self._selector_layout)

        self._selectors = dict()

    @property
    def selection(self) -> Dict[str, bool]:
        return {name: selector.isChecked() for name, (_, selector) in self._selectors.items()}

    def set_selection(self, selection: Mapping[str, bool]) -> None:
        for name, select in selection.items():
            try:
                self._selectors[name][1].setChecked(select)
            except KeyError:
                pass

    def add_selector(self,
                     name: str,
                     item: Optional[pg.PlotDataItem] = None,
                     selected: Optional[bool] = False) -> None:
        if name in self._selectors:
            raise ValueError(f'Selector with name "{name}" already present in plot selector')
        selector = self._create_selector(name)
        selector.setChecked(selected)
        selector.clicked.connect(self._selection_changed)
        self._selector_layout.removeItem(self._stretch)
        row = len(self._selectors)
        if item is None:
            self._selector_layout.addWidget(selector, row, 0, 1, 2)
            self._selectors[name] = (None, selector)
        else:
            icon = PlotLegendIconWidget(item)
            self._selector_layout.addWidget(icon, row, 0)
            self._selector_layout.addWidget(selector, row, 1)
            self._selectors[name] = (icon, selector)
        self._selector_layout.addItem(self._stretch, row + 1, 0, 1, 2)

    def remove_selector(self, name: str) -> None:
        if name not in self._selectors:
            raise ValueError(f'Selector with name "{name}" not found in plot selector')
        self._selector_layout.removeItem(self._stretch)
        for sel_name, (icon, selector) in reversed(self._selectors.items()):
            self._selector_layout.removeWidget(selector)
            if icon is not None:
                self._selector_layout.removeWidget(icon)
            if sel_name == name:
                break
        after_remove = False
        for row, (sel_name, (icon, selector)) in enumerate(self._selectors.items()):
            if after_remove:
                if icon is None:
                    self._selector_layout.addWidget(selector, row - 1, 0, 1, 2)
                else:
                    self._selector_layout.addWidget(icon, row - 1, 0)
                    self._selector_layout.addWidget(selector, row - 1, 1)
            elif sel_name == name:
                after_remove = True
        icon, selector = self._selectors.pop(name)
        selector.clicked.disconnect()
        selector.setParent(None)
        icon.setParent(None)
        self._selector_layout.addItem(self._stretch, len(self._selectors), 0, 1, 2)

    def _selection_changed(self) -> None:
        self.sigSelectionChanged.emit(self.selection)

    @staticmethod
    def _create_selector(name: str, color: Optional[Any] = None) -> QtWidgets.QCheckBox:
        checkbox = QtWidgets.QCheckBox(name)
        if color is not None:
            color_str = pg.mkColor(color).name()
            checkbox.setStyleSheet('QCheckBox { color: ' + color_str + ' }')
        return checkbox


class CursorPositionLabel(QtWidgets.QLabel):
    """
    """

    def __init__(self,
                 units: Optional[Tuple[str, str]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        self._units = ('', '')
        self._text_template = ''
        self._pos_cache = (0, 0)

        if units is None:
            units = self._units
        self.set_units(*units)

    def set_units(self, x: str, y: str) -> None:
        units = (x if x else '', y if y else '')
        self._update_text_template(units)
        self._units = units
        self.update_position(self._pos_cache)

    def update_position(self, pos: Tuple[float, float]) -> None:
        x = ScaledFloat(pos[0])
        y = ScaledFloat(pos[1])
        self.setText(self._text_template.format(x, y))
        self._pos_cache = pos

    def _update_text_template(self, units: Tuple[str, str]) -> None:
        x_unit, y_unit = units
        self._text_template = f'Cursor: ({{:.3r}}{x_unit}, {{:.3r}}{y_unit})'


class InteractiveCurvesWidget(QtWidgets.QWidget):
    """
    """

    SelectionMode = RubberbandZoomSelectionPlotWidget.SelectionMode
    
    sigPlotParametersChanged = QtCore.Signal()
    sigAutoLimitsApplied = QtCore.Signal(bool, bool)  # in x- and/or y-direction

    def __init__(self,
                 allow_tracking_outside_data: Optional[bool] = False,
                 max_mouse_pos_update_rate: Optional[float] = None,
                 selection_bounds: Optional[Sequence[Tuple[Union[None, float], Union[None, float]]]] = None,
                 selection_pen: Optional[Any] = None,
                 selection_hover_pen: Optional[Any] = None,
                 selection_brush: Optional[Any] = None,
                 selection_hover_brush: Optional[Any] = None,
                 xy_region_selection_crosshair: Optional[bool] = False,
                 xy_region_selection_handles: Optional[bool] = True,
                 **kwargs
                 ) -> None:
        super().__init__(**kwargs)
        if max_mouse_pos_update_rate is None:
            max_mouse_pos_update_rate = 20.

        self._plot_widget = PlotWidget(
            allow_tracking_outside_data=allow_tracking_outside_data,
            max_mouse_pos_update_rate=max_mouse_pos_update_rate,
            selection_bounds=selection_bounds,
            selection_pen=selection_pen,
            selection_hover_pen=selection_hover_pen,
            selection_brush=selection_brush,
            selection_hover_brush=selection_hover_brush,
            xy_region_selection_crosshair=xy_region_selection_crosshair,
            xy_region_selection_handles=xy_region_selection_handles,
        )
        self._plot_legend = self._plot_widget.addLegend()
        self._plot_legend.hide()
        self._plot_editor = PlotEditorWidget()
        self._plot_selector = PlotSelectorWidget()
        self._position_label = CursorPositionLabel(units=self._plot_editor.units)

        self._plot_editor.layout().setContentsMargins(0, 0, 0, 0)
        self._plot_selector.layout().setContentsMargins(0, 0, 0, 0)

        layout = QtWidgets.QGridLayout()
        layout.addWidget(self._position_label, 0, 0)
        layout.addWidget(self._plot_widget, 1, 0)
        layout.addWidget(self._plot_editor, 2, 0, 1, 2)
        layout.addWidget(self._plot_selector, 1, 1)
        layout.setColumnStretch(0, 1)
        layout.setRowStretch(1, 1)
        self.setLayout(layout)

        self._plot_selector.sigSelectionChanged.connect(self._update_plot_selection)
        self._plot_editor.sigUnitsChanged.connect(self.__units_changed)
        self._plot_editor.sigLabelsChanged.connect(self.__labels_changed)
        self._plot_editor.sigLimitsChanged.connect(self.__limits_changed)
        self._plot_editor.sigAutoRangeClicked.connect(self.set_auto_range)
        self._plot_widget.sigRangeChanged.connect(self.__plot_widget_limits_changed)
        self._plot_widget.sigMouseMoved.connect(self._position_label.update_position)

        self.__labels_changed(*self.labels)
        self.__units_changed(*self.units)
        self.set_auto_range(True, True)

        # patch attributes of advanced PlotWidget into this widget for easier access and
        # auto-completion
        self.set_rubberband_zoom_selection_mode = self._plot_widget.set_rubberband_zoom_selection_mode
        self.set_region_selection_mode = self._plot_widget.set_region_selection_mode
        self.set_marker_selection_mode = self._plot_widget.set_marker_selection_mode
        self.set_selection_mutable = self._plot_widget.set_selection_mutable
        self.set_selection_bounds = self._plot_widget.set_selection_bounds
        self.add_region_selection = self._plot_widget.add_region_selection
        self.add_marker_selection = self._plot_widget.add_marker_selection
        self.move_region_selection = self._plot_widget.move_region_selection
        self.move_marker_selection = self._plot_widget.move_marker_selection
        self.clear_marker_selections = self._plot_widget.clear_marker_selections
        self.delete_marker_selection = self._plot_widget.delete_marker_selection
        self.hide_marker_selections = self._plot_widget.hide_marker_selections
        self.show_marker_selections = self._plot_widget.show_marker_selections
        self.hide_marker_selection = self._plot_widget.hide_marker_selection
        self.show_marker_selection = self._plot_widget.show_marker_selection
        self.clear_region_selections = self._plot_widget.clear_region_selections
        self.delete_region_selection = self._plot_widget.delete_region_selection
        self.hide_region_selections = self._plot_widget.hide_region_selections
        self.show_region_selections = self._plot_widget.show_region_selections
        self.hide_region_selection = self._plot_widget.hide_region_selection
        self.show_region_selection = self._plot_widget.show_region_selection

        # Disable bugged pyqtgraph interactive mouse menu options to avoid a myriad of
        # user-induced errors.
        for action in self._plot_widget.getPlotItem().ctrlMenu.actions():
            if action.text() not in ('Alpha', 'Grid', 'Points'):
                action.setEnabled(False)
                action.setVisible(False)
        for axis_ctrl in self._plot_widget.getViewBox().menu.ctrl:
            axis_ctrl.autoPanCheck.setEnabled(False)
            axis_ctrl.visibleOnlyCheck.setEnabled(False)
            axis_ctrl.linkCombo.setEnabled(False)
            axis_ctrl.label.setEnabled(False)
            axis_ctrl.autoPanCheck.setVisible(False)
            axis_ctrl.visibleOnlyCheck.setVisible(False)
            axis_ctrl.linkCombo.setVisible(False)
            axis_ctrl.label.setVisible(False)

        # Keep track of PlotItems plotted
        self._plot_items = dict()
        self._fit_plot_items = dict()

    def _get_valid_generic_name(self, index: Optional[int] = 1) -> str:
        name = f'Dataset {index:d}'
        if name in self._plot_items:
            return self._get_valid_generic_name(index + 1)
        return name

    def plot(self, name: Optional[str] = None, **kwargs) -> str:
        # Delete old plot if present
        if name is None:
            name = self._get_valid_generic_name()
        elif name in self._plot_items:
            self.remove_plot(name)
        # Add new plot and enable antialias by default if not explicitly set
        antialias = kwargs.pop('antialias', True)
        item = self._plot_widget.plot(name=name, antialias=antialias, **kwargs)
        self._plot_items[name] = item
        self._plot_selector.add_selector(name=name, item=item, selected=True)
        return name

    def remove_plot(self, name: str) -> None:
        self.remove_fit_plot(name)
        item = self._plot_items.pop(name, None)
        if item in self._plot_widget.getViewBox().addedItems:
            self._plot_widget.removeItem(item)
        try:
            self._plot_selector.remove_selector(name)
        except ValueError:
            pass

    def clear(self) -> None:
        for name in list(self._plot_items):
            self.remove_plot(name)

    def clear_fits(self) -> None:
        for name in list(self._fit_plot_items):
            self.remove_fit_plot(name)

    def plot_fit(self, name: str, **kwargs) -> None:
        if name not in self._plot_items:
            raise ValueError(f'No plot with name "{name}" found to add fit to')
        # Delete old plot if present
        if name in self._fit_plot_items:
            self.remove_fit_plot(name)
        # Add new plot and enable antialias by default if not explicitly set
        antialias = kwargs.pop('antialias', True)
        item = self._plot_widget.plot(name=None, antialias=antialias, **kwargs)
        self._fit_plot_items[name] = item

    def remove_fit_plot(self, name: str) -> None:
        item = self._fit_plot_items.pop(name, None)
        if item in self._plot_widget.getViewBox().addedItems:
            self._plot_widget.removeItem(item)

    def set_data(self, name: str, *args, **kwargs) -> None:
        """ See pyqtgraph.PlotDataItem.__init__ for valid arguments """
        self._plot_items[name].setData(*args, **kwargs)

    def set_fit_data(self, name: str, *args, **kwargs) -> None:
        """ See pyqtgraph.PlotDataItem.__init__ for valid arguments """
        if name not in self._fit_plot_items:
            self.plot_fit(name)
        self._fit_plot_items[name].setData(*args, **kwargs)

    @property
    def plot_names(self) -> List[str]:
        return list(self._plot_items)

    @property
    def plot_selection(self) -> Dict[str, bool]:
        return {name: item.isVisible() for name, item in self._plot_items.items()}

    def set_plot_selection(self, selection: Mapping[str, bool]) -> None:
        self._plot_selector.set_selection(selection)
        self._update_plot_selection(selection)

    def set_auto_range(self, x: Optional[bool] = None, y: Optional[bool] = None) -> None:
        if x is y is None:
            return
        if x is not None:
            self._plot_widget.enableAutoRange(axis='x', enable=x)
        if y is not None:
            self._plot_widget.enableAutoRange(axis='y', enable=y)
        self.sigAutoLimitsApplied.emit(bool(x), bool(y))

    def set_labels(self, x: Optional[str] = None, y: Optional[str] = None) -> None:
        self._plot_editor.set_labels(x, y)
        self.__labels_changed(*self.labels)

    def set_units(self, x: Optional[str] = None, y: Optional[str] = None) -> None:
        self._plot_editor.set_units(x, y)
        self.__units_changed(*self.units)

    def set_limits(self,
                   x: Optional[Tuple[float, float]] = None,
                   y: Optional[Tuple[float, float]] = None
                   ) -> None:
        self._plot_editor.set_limits(x, y)
        self.__limits_changed(*self.limits)

    # Start of attribute/property wrapping of sub-widgets

    @property
    def sigMarkerSelectionChanged(self) -> QtCore.Signal:
        return self._plot_widget.sigMarkerSelectionChanged

    @property
    def sigRegionSelectionChanged(self) -> QtCore.Signal:
        return self._plot_widget.sigRegionSelectionChanged

    @property
    def sigMouseMoved(self) -> QtCore.Signal:
        return self._plot_widget.sigMouseMoved

    @property
    def sigMouseDragged(self) -> QtCore.Signal:
        return self._plot_widget.sigMouseDragged

    @property
    def sigMouseClicked(self) -> QtCore.Signal:
        return self._plot_widget.sigMouseClicked

    @property
    def sigZoomAreaApplied(self) -> QtCore.Signal:
        return self._plot_widget.sigZoomAreaApplied

    @property
    def labels(self) -> Tuple[str, str]:
        return self._plot_editor.labels

    @property
    def units(self) -> Tuple[str, str]:
        return self._plot_editor.units

    @property
    def limits(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        return self._plot_editor.limits

    @property
    def rubberband_zoom_selection_mode(self) -> SelectionMode:
        return self._plot_widget.rubberband_zoom_selection_mode

    @property
    def marker_selection(self) -> Dict[SelectionMode, List[Union[float, Tuple[float, float]]]]:
        return self._plot_widget.marker_selection

    @property
    def region_selection(self) -> Dict[SelectionMode, List[tuple]]:
        return self._plot_widget.region_selection

    @property
    def region_selection_mode(self) -> SelectionMode:
        return self._plot_widget.region_selection_mode

    @property
    def marker_selection_mode(self) -> SelectionMode:
        return self._plot_widget.marker_selection_mode

    @property
    def selection_mutable(self) -> bool:
        return self._plot_widget.selection_mutable

    @property
    def selection_bounds(self) -> Union[None, List[Union[None, Tuple[float, float]]]]:
        return self._plot_widget.selection_bounds

    @property
    def allow_tracking_outside_data(self) -> bool:
        return self._plot_widget.allow_tracking_outside_data

    @allow_tracking_outside_data.setter
    def allow_tracking_outside_data(self, allow: bool) -> None:
        self._plot_widget.allow_tracking_outside_data = bool(allow)

    # Start of methods to show/hide sub-widgets

    def toggle_plot_selector(self, enable: bool) -> None:
        # Sync legend and selector checkboxes
        if not self._plot_selector.isVisible() and enable:
            self._plot_selector.set_selection(self.plot_selection)
        self._plot_selector.setVisible(enable)
        self._plot_legend.setVisible(not enable)

    def toggle_plot_editor(self, enable: bool) -> None:
        self._plot_editor.setVisible(enable)

    def toggle_cursor_position(self, enable: bool) -> None:
        is_enabled = self._position_label.isVisible()
        self._position_label.setVisible(enable)
        if is_enabled and not enable:
            self._plot_widget.sigMouseMoved.disconnect(self._position_label.update_position)
        elif not is_enabled and enable:
            self._plot_widget.sigMouseMoved.connect(self._position_label.update_position)

    # Start of slots for internal updates

    def _update_plot_selection(self, selection: Mapping[str, bool]) -> None:
        for name, selected in selection.items():
            try:
                self._plot_items[name].setVisible(selected)
            except KeyError:
                pass
            else:
                try:
                    self._fit_plot_items[name].setVisible(selected)
                except KeyError:
                    pass

    def __units_changed(self, x: Optional[str] = None, y: Optional[str] = None) -> None:
        if x is y is None:
            return
        x_label, y_label = self.labels
        if x is not None:
            self._plot_widget.setLabel('bottom', x_label, units=x)
        if y is not None:
            self._plot_widget.setLabel('left', y_label, units=y)
        self._position_label.set_units(*self.units)
        self.sigPlotParametersChanged.emit()

    def __labels_changed(self, x: Optional[str] = None, y: Optional[str] = None) -> None:
        if x is y is None:
            return
        x_unit, y_unit = self.units
        if x is not None:
            self._plot_widget.setLabel('bottom', x, units=x_unit)
        if y is not None:
            self._plot_widget.setLabel('left', y, units=y_unit)
        self.sigPlotParametersChanged.emit()

    def __limits_changed(self,
                         x: Optional[Tuple[float, float]] = None,
                         y: Optional[Tuple[float, float]] = None
                         ) -> None:
        if x is y is None:
            return
        if x is not None:
            self._plot_widget.enableAutoRange(axis='x', enable=False)
            self._plot_widget.setXRange(*x, padding=0)
        if y is not None:
            self._plot_widget.enableAutoRange(axis='y', enable=False)
            self._plot_widget.setYRange(*y, padding=0)
        # Signal is emitted once the pyqtgraph plot has actually changed.
        # See: self.__plot_widget_limits_changed

    def __plot_widget_limits_changed(self,
                                     _,
                                     limits: Tuple[Tuple[float, float], Tuple[float, float]]
                                     ) -> None:
        self._plot_editor.set_limits(*limits)
        self.sigPlotParametersChanged.emit()



# -*- coding: utf-8 -*-

"""
This file contains modified pyqtgraph plot widgets for advanced interactive plotting.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['PlotWidget', 'MouseTrackingPlotWidget', 'RubberbandZoomPlotWidget',
           'DataSelectionPlotWidget', 'RubberbandZoomSelectionPlotWidget', 'MouseTrackingMixin',
           'RubberbandZoomMixin', 'DataSelectionMixin']

from typing import Union, Tuple, List, Dict, Optional, Any, Sequence
from PySide2 import QtCore
from pyqtgraph import PlotWidget as _PlotWidget
from pyqtgraph import SignalProxy as _SignalProxy
import qudi.util.widgets.plotting.view_box as _vb


class MouseTrackingMixin:
    """ Extend the PlotWidget class with mouse tracking and signalling """

    # position (x, y)
    sigMouseMoved = QtCore.Signal(tuple)

    def __init__(self,
                 allow_tracking_outside_data: Optional[bool] = False,
                 max_mouse_pos_update_rate: Optional[float] = None,
                 **kwargs
                 ) -> None:
        if not isinstance(kwargs.get('viewBox', None), _vb.MouseTrackingMixin):
            # Use custom pg.ViewBox subclass
            kwargs['viewBox'] = _vb.MouseTrackingViewBox(
                allow_tracking_outside_data=allow_tracking_outside_data
            )

        super().__init__(**kwargs)

        if max_mouse_pos_update_rate is not None and max_mouse_pos_update_rate > 0.:
            self._mouse_position_signal_proxy = _SignalProxy(
                signal=self.scene().sigMouseMoved,
                rateLimit=max_mouse_pos_update_rate,
                delay=2 / max_mouse_pos_update_rate,  # Must be larger than 1/rateLimit
                slot=self.__mouse_moved
            )

    def __mouse_moved(self, args) -> None:
        pos = self.getViewBox().mapSceneToView(args[0])
        self.sigMouseMoved.emit((pos.x(), pos.y()))

    @property
    def sigMouseDragged(self) -> QtCore.Signal:
        return self.getViewBox().sigMouseDragged

    @property
    def sigMouseClicked(self) -> QtCore.Signal:
        return self.getViewBox().sigMouseClicked


class RubberbandZoomMixin:

    SelectionMode = _vb.SelectionMode

    def __init__(self, **kwargs):
        if not isinstance(kwargs.get('viewBox', None), _vb.RubberbandZoomMixin):
            # Use custom pg.ViewBox subclass
            kwargs['viewBox'] = _vb.RubberbandZoomViewBox()
        super().__init__(**kwargs)
        self.set_rubberband_zoom_selection_mode = self.getViewBox().set_rubberband_zoom_selection_mode

    @property
    def sigZoomAreaApplied(self) -> QtCore.Signal:
        return self.getViewBox().sigZoomAreaApplied

    @property
    def rubberband_zoom_selection_mode(self) -> SelectionMode:
        return self.getViewBox().rubberband_zoom_selection_mode


class DataSelectionMixin:
    """ Extend the PlotWidget class with mouse tracking and signalling as well as mouse pointer
    data selection tools.
    """
    SelectionMode = _vb.SelectionMode

    def __init__(self,
                 selection_bounds: Optional[Sequence[Tuple[Union[None, float], Union[None, float]]]] = None,
                 selection_pen: Optional[Any] = None,
                 selection_hover_pen: Optional[Any] = None,
                 selection_brush: Optional[Any] = None,
                 selection_hover_brush: Optional[Any] = None,
                 xy_region_selection_crosshair: Optional[bool] = False,
                 xy_region_selection_handles: Optional[bool] = True,
                 xy_region_min_size_percentile: Optional[float] = None,
                 **kwargs
                 ) -> None:
        if not isinstance(kwargs.get('viewBox', None), _vb.DataSelectionMixin):
            # Use custom pg.ViewBox subclass
            kwargs['viewBox'] = _vb.DataSelectionViewBox(
                selection_bounds=selection_bounds,
                selection_pen=selection_pen,
                selection_hover_pen=selection_hover_pen,
                selection_brush=selection_brush,
                selection_hover_brush=selection_hover_brush,
                xy_region_selection_crosshair=xy_region_selection_crosshair,
                xy_region_selection_handles=xy_region_selection_handles,
                xy_region_min_size_percentile=xy_region_min_size_percentile
            )
        super().__init__(**kwargs)
        vb = self.getViewBox()
        self.set_region_selection_mode = vb.set_region_selection_mode
        self.set_marker_selection_mode = vb.set_marker_selection_mode
        self.set_selection_mutable = vb.set_selection_mutable
        self.set_selection_bounds = vb.set_selection_bounds
        self.add_region_selection = vb.add_region_selection
        self.add_marker_selection = vb.add_marker_selection
        self.move_region_selection = vb.move_region_selection
        self.move_marker_selection = vb.move_marker_selection
        self.clear_marker_selections = vb.clear_marker_selections
        self.delete_marker_selection = vb.delete_marker_selection
        self.clear_region_selections = vb.clear_region_selections
        self.delete_region_selection = vb.delete_region_selection
        self.hide_marker_selections = vb.hide_marker_selections
        self.show_marker_selections = vb.show_marker_selections
        self.hide_marker_selection = vb.hide_marker_selection
        self.show_marker_selection = vb.show_marker_selection
        self.hide_region_selections = vb.hide_region_selections
        self.show_region_selections = vb.show_region_selections
        self.hide_region_selection = vb.hide_region_selection
        self.show_region_selection = vb.show_region_selection

    @property
    def sigMarkerSelectionChanged(self) -> QtCore.Signal:
        return self.getViewBox().sigMarkerSelectionChanged

    @property
    def sigRegionSelectionChanged(self) -> QtCore.Signal:
        return self.getViewBox().sigRegionSelectionChanged

    @property
    def marker_selection(self) -> Dict[SelectionMode, List[Union[float, Tuple[float, float]]]]:
        return self.getViewBox().marker_selection

    @property
    def region_selection(self) -> Dict[SelectionMode, List[Tuple[Tuple[float, float], Tuple[float, float]]]]:
        return self.getViewBox().region_selection

    @property
    def region_selection_mode(self) -> SelectionMode:
        return self.getViewBox().region_selection_mode

    @property
    def marker_selection_mode(self) -> SelectionMode:
        return self.getViewBox().marker_selection_mode

    @property
    def selection_mutable(self) -> bool:
        return self.getViewBox().selection_mutable

    @property
    def selection_bounds(self) -> Union[None, List[Union[None, Tuple[float, float]]]]:
        return self.getViewBox().selection_bounds


class PlotWidget(_PlotWidget):
    """ Make blockSignals also un-/mute signals from the viewbox """
    def blockSignals(self, block: bool) -> None:
        super().blockSignals(block)
        self.getViewBox().blockSignals(block)


class MouseTrackingPlotWidget(MouseTrackingMixin, PlotWidget):
    """ Extend the PlotWidget class with mouse tracking and signalling """
    pass


class RubberbandZoomPlotWidget(RubberbandZoomMixin, MouseTrackingMixin, PlotWidget):
    """ Extend the PlotWidget class with mouse tracking and signalling as well as a rubberband zoom
    tool.
    """
    pass


class DataSelectionPlotWidget(DataSelectionMixin, MouseTrackingMixin, PlotWidget):
    """ Extend the PlotWidget class with mouse tracking and signalling as well as mouse pointer
    data selection tools.
    """
    pass


class RubberbandZoomSelectionPlotWidget(RubberbandZoomMixin, DataSelectionMixin, MouseTrackingMixin, PlotWidget):
    """ Extend the PlotWidget class with mouse tracking and signalling as well as mouse pointer
    data selection tools and rubberband zoom feature.
    """
    def __init__(self,
                 allow_tracking_outside_data: Optional[bool] = False,
                 selection_bounds: Optional[Sequence[Tuple[Union[None, float], Union[None, float]]]] = None,
                 selection_pen: Optional[Any] = None,
                 selection_hover_pen: Optional[Any] = None,
                 selection_brush: Optional[Any] = None,
                 selection_hover_brush: Optional[Any] = None,
                 xy_region_selection_crosshair: Optional[bool] = False,
                 xy_region_selection_handles: Optional[bool] = True,
                 xy_region_min_size_percentile: Optional[float] = None,
                 **kwargs
                 ) -> None:
        has_selection = isinstance(kwargs.get('viewBox', None), _vb.DataSelectionMixin)
        has_rubberband = isinstance(kwargs.get('viewBox', None), _vb.RubberbandZoomMixin)
        if not has_selection or not has_rubberband:
            kwargs['viewBox'] = _vb.RubberbandZoomSelectionViewBox(
                allow_tracking_outside_data=allow_tracking_outside_data,
                selection_bounds=selection_bounds,
                selection_pen=selection_pen,
                selection_hover_pen=selection_hover_pen,
                selection_brush=selection_brush,
                selection_hover_brush=selection_hover_brush,
                xy_region_selection_crosshair=xy_region_selection_crosshair,
                xy_region_selection_handles=xy_region_selection_handles,
                xy_region_min_size_percentile=xy_region_min_size_percentile
            )
        super().__init__(**kwargs)



# -*- coding: utf-8 -*-

"""
This file contains custom widgets to facilitate 2D image display with an adjustable colorscale and
various other interactive features.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ImageWidget', 'MouseTrackingImageWidget', 'RubberbandZoomImageWidget',
           'DataSelectionImageWidget', 'RubberbandZoomSelectionImageWidget']

from typing import Union, Optional, Tuple, List, Dict
from PySide2 import QtCore, QtWidgets
from pyqtgraph import PlotWidget as _PlotWidget
from qudi.util.widgets.plotting.plot_item import DataImageItem as _DataImageItem
from qudi.util.widgets.plotting.colorbar import ColorBarWidget as _ColorBarWidget
import qudi.util.widgets.plotting.plot_widget as _pw


class ImageWidget(QtWidgets.QWidget):
    """ Composite widget consisting of a PlotWidget and a colorbar to display 2D image data.
    Provides a convenient image data interface and handles user colorscale interaction.
    """
    _plot_widget_type = _PlotWidget

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None, **kwargs):
        super().__init__(parent=parent)

        self.plot_widget = self._plot_widget_type(**kwargs)
        self.image_item = _DataImageItem()
        self.plot_widget.addItem(self.image_item)
        self.plot_widget.setAspectLocked(lock=True, ratio=1.0)

        self.colorbar_widget = _ColorBarWidget()
        if self.colorbar_widget.mode == _ColorBarWidget.ColorBarMode.PERCENTILE:
            self.image_item.set_percentiles(self.colorbar_widget.percentiles)
        else:
            self.image_item.set_percentiles(None)
        self.colorbar_widget.sigModeChanged.connect(self._colorbar_mode_changed)
        self.colorbar_widget.sigLimitsChanged.connect(self._colorbar_limits_changed)
        self.colorbar_widget.sigPercentilesChanged.connect(self._colorbar_percentiles_changed)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setStretch(0, 1)
        layout.addWidget(self.plot_widget)
        layout.addWidget(self.colorbar_widget)
        self.setLayout(layout)

        # monkey-patch attributes from PlotWidget and ImageItem for convenient access
        self.autoRange = self.plot_widget.autoRange
        self.set_percentiles = self.image_item.set_percentiles
        self.set_image_extent = self.image_item.set_image_extent

    @property
    def percentiles(self) -> Union[None, Tuple[float, float]]:
        return self.image_item.percentiles

    @property
    def levels(self) -> Tuple[float, float]:
        return self.image_item.levels

    def set_image(self,
                  image,
                  extent: Optional[Tuple[Tuple[float, float], Tuple[float, float]]] = None,
                  adjust_for_px_size: Optional[bool] = True
                  ) -> None:
        """
        """
        if image is None:
            self.image_item.set_image(image=None, autoLevels=False)
            return

        # Set image with proper colorbar limits
        if self.colorbar_widget.mode == _ColorBarWidget.ColorBarMode.PERCENTILE:
            self.image_item.set_image(image=image, autoLevels=False)
            levels = self.image_item.levels
            if levels is not None:
                self.colorbar_widget.set_limits(*levels)
        else:
            self.image_item.set_image(image=image,
                                      autoLevels=False,
                                      levels=self.colorbar_widget.limits)
        if extent is not None:
            self.image_item.set_image_extent(extent, adjust_for_px_size)

    def set_axis_label(self, axis, label=None, unit=None):
        return self.plot_widget.setLabel(axis, text=label, units=unit)

    def set_data_label(self, label, unit=None):
        return self.colorbar_widget.set_label(label, unit)

    def _colorbar_mode_changed(self, mode: _ColorBarWidget.ColorBarMode) -> None:
        if mode == _ColorBarWidget.ColorBarMode.PERCENTILE:
            self._colorbar_percentiles_changed(self.colorbar_widget.percentiles)
        else:
            self._colorbar_limits_changed(self.colorbar_widget.limits)

    def _colorbar_limits_changed(self, limits: Tuple[float, float]) -> None:
        self.image_item.set_percentiles(None)
        self.image_item.setLevels(limits)

    def _colorbar_percentiles_changed(self, percentiles: Tuple[float, float]) -> None:
        self.image_item.set_percentiles(percentiles)
        levels = self.levels
        if levels is not None:
            self.colorbar_widget.set_limits(*levels)


class MouseTrackingMixin:
    """ Extends the normal qudi ImageWidget with a custom PlotWidget type that tracks mouse
    activity and sends signals.
    """
    _plot_widget_type = _pw.MouseTrackingPlotWidget

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None, **kwargs):
        super().__init__(parent=parent, **kwargs)

    @property
    def sigMouseMoved(self) -> QtCore.Signal:
        return self.plot_widget.sigMouseMoved

    @property
    def sigMouseDragged(self) -> QtCore.Signal:
        return self.plot_widget.sigMouseDragged

    @property
    def sigMouseClicked(self) -> QtCore.Signal:
        return self.plot_widget.sigMouseClicked


class RubberbandZoomMixin:
    """ Extends the qudi MouseTrackingImageWidget with a rubberband zoom tool.
    """
    _plot_widget_type = _pw.RubberbandZoomPlotWidget
    SelectionMode = _pw.RubberbandZoomPlotWidget.SelectionMode

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.set_rubberband_zoom_selection_mode = self.plot_widget.set_rubberband_zoom_selection_mode

    @property
    def sigZoomAreaApplied(self) -> QtCore.Signal:
        return self.plot_widget.sigZoomAreaApplied

    @property
    def rubberband_zoom_selection_mode(self) -> SelectionMode:
        return self.plot_widget.rubberband_zoom_selection_mode


class DataSelectionMixin:
    """ Extends the qudi MouseTrackingImageWidget with data selection tools and signals.
    """
    _plot_widget_type = _pw.DataSelectionPlotWidget
    SelectionMode = _pw.DataSelectionPlotWidget.SelectionMode

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.set_region_selection_mode = self.plot_widget.set_region_selection_mode
        self.set_marker_selection_mode = self.plot_widget.set_marker_selection_mode
        self.set_selection_mutable = self.plot_widget.set_selection_mutable
        self.set_selection_bounds = self.plot_widget.set_selection_bounds
        self.add_region_selection = self.plot_widget.add_region_selection
        self.add_marker_selection = self.plot_widget.add_marker_selection
        self.move_region_selection = self.plot_widget.move_region_selection
        self.move_marker_selection = self.plot_widget.move_marker_selection
        self.clear_marker_selections = self.plot_widget.clear_marker_selections
        self.delete_marker_selection = self.plot_widget.delete_marker_selection
        self.clear_region_selections = self.plot_widget.clear_region_selections
        self.delete_region_selection = self.plot_widget.delete_region_selection
        self.hide_marker_selections = self.plot_widget.hide_marker_selections
        self.show_marker_selections = self.plot_widget.show_marker_selections
        self.hide_marker_selection = self.plot_widget.hide_marker_selection
        self.show_marker_selection = self.plot_widget.show_marker_selection
        self.hide_region_selections = self.plot_widget.hide_region_selections
        self.show_region_selections = self.plot_widget.show_region_selections
        self.hide_region_selection = self.plot_widget.hide_region_selection
        self.show_region_selection = self.plot_widget.show_region_selection

    @property
    def sigMarkerSelectionChanged(self) -> QtCore.Signal:
        return self.plot_widget.sigMarkerSelectionChanged

    @property
    def sigRegionSelectionChanged(self) -> QtCore.Signal:
        return self.plot_widget.sigRegionSelectionChanged

    @property
    def marker_selection(self) -> Dict[SelectionMode, List[Tuple[float, float]]]:
        return self.plot_widget.marker_selection

    @property
    def region_selection(self) -> Dict[SelectionMode, List[Tuple[Tuple[float, float], Tuple[float, float]]]]:
        return self.plot_widget.region_selection

    @property
    def region_selection_mode(self) -> _pw.DataSelectionPlotWidget.SelectionMode:
        return self.plot_widget.region_selection_mode

    @property
    def marker_selection_mode(self) -> _pw.DataSelectionPlotWidget.SelectionMode:
        return self.plot_widget.marker_selection_mode

    @property
    def selection_mutable(self) -> bool:
        return self.plot_widget.selection_mutable

    @property
    def selection_bounds(self) -> Union[None, List[Union[None, Tuple[float, float]]]]:
        return self.plot_widget.selection_bounds


class MouseTrackingImageWidget(MouseTrackingMixin, ImageWidget):
    pass


class RubberbandZoomImageWidget(RubberbandZoomMixin, MouseTrackingMixin, ImageWidget):
    pass


class DataSelectionImageWidget(DataSelectionMixin, MouseTrackingMixin, ImageWidget):
    pass


class RubberbandZoomSelectionImageWidget(RubberbandZoomMixin, DataSelectionMixin, MouseTrackingMixin, ImageWidget):
    _plot_widget_type = _pw.RubberbandZoomSelectionPlotWidget
    SelectionMode = _pw.RubberbandZoomSelectionPlotWidget.SelectionMode



# -*- coding: utf-8 -*-
"""
Parent poller mechanism from IPython.

Copyright (c) 2015, IPython Development Team

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ParentPollerUnix', 'ParentPollerWindows']

import ctypes
import os
import platform
import time
from threading import Thread
import logging

logger = logging.getLogger(__name__)


class ParentPollerUnix(Thread):
    """A Unix-specific daemon thread that terminates the program immediately
    when the parent process no longer exists.
    """

    def __init__(self, quit_function=None):
        """Create the parent poller.

        Parameters
        ----------
        quitfunction : callable
            Function to run before exiting.
        """
        if quit_function is None:
            pass
        elif not callable(quit_function):
            raise TypeError('argument quit_function must be a callable.')
        super().__init__()
        self.daemon = True
        self.quit_function = quit_function

    def run(self):
        """Run the parentpoller.
        """
        # We cannot use os.waitpid because it works only for child processes.
        from errno import EINTR
        while True:
            try:
                if os.getppid() == 1:
                    if self.quit_function is None:
                        logger.critical('Parent process died!')
                    else:
                        logger.critical('Parent process died! Qudi shutting down...')
                        self.quit_function()
                    return
            except OSError as e:
                if e.errno == EINTR:
                    continue
                raise
            time.sleep(1)


class ParentPollerWindows(Thread):
    """A Windows-specific daemon thread that listens for a special event that signals an interrupt
    and, optionally, terminates the program immediately when the parent process no longer exists.
    """

    def __init__(self, parent_handle, quit_function=None):
        """ Create the parent poller.

        Parameters
        ----------
        quit_function : callable
            Function to call for shutdown if the parent process is dead.
        parent_handle : int
            The program will terminate immediately when this handle is signaled.
        """
        if quit_function is None:
            pass
        elif not callable(quit_function):
            raise TypeError('argument quit_function must be a callable.')
        super().__init__()
        self.daemon = True
        self.quit_function = quit_function
        self.parent_handle = parent_handle
        self._stop_requested = False

    def run(self):
        """Run the poll loop. This method never returns.
        """
        try:
            from _winapi import WAIT_OBJECT_0, INFINITE
        except ImportError:
            from _subprocess import WAIT_OBJECT_0, INFINITE

        # Build the list of handle to listen on.
        handle_list = [self.parent_handle]
        arch = platform.architecture()[0]
        c_int = ctypes.c_int64 if arch.startswith('64') else ctypes.c_int

        # Listen forever.
        while True:
            # Return if stop has been requested
            if self._stop_requested:
                return

            result = ctypes.windll.kernel32.WaitForMultipleObjects(
                len(handle_list),                           # nCount
                (c_int * len(handle_list))(*handle_list),   # lpHandles
                False,                                      # bWaitAll
                1000)                                       # dwMilliseconds

            if result >= len(handle_list):
                # Nothing happened. Probably timed out.
                continue
            elif result < WAIT_OBJECT_0:
                # wait failed, just give up and stop polling.
                logger.critical("Parent poll failed!!!!!")
                return
            else:
                handle = handle_list[result - WAIT_OBJECT_0]
                if handle == self.parent_handle:
                    if self.quit_function is None:
                        logger.critical('Parent process died!')
                    else:
                        logger.critical('Parent process died! Qudi shutting down...')
                        self.quit_function()
                    return

    def stop(self) -> None:
        self._stop_requested = True

    def start(self) -> None:
        self._stop_requested = False
        return super().start()



# -*- coding: utf-8 -*-
"""
This file contains the Qudi task runner module.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from functools import partial
from PySide2 import QtCore
from typing import Any, Type, Mapping, List, Dict

from qudi.util.mutex import Mutex
from qudi.core.module import LogicBase
from qudi.core.scripting.moduletask import ModuleTask
from qudi.core.scripting.modulescript import import_module_script
from qudi.core.configoption import ConfigOption


class TaskRunnerLogic(LogicBase):
    """This module keeps a collection of available ModuleTask subclasses (defined by config) and
    respective initialized instances that can be run.
    Handles module connections to tasks and allows monitoring of task states and results.
    """

    _module_task_configs = ConfigOption(name='module_tasks', default=dict(), missing='warn')

    sigTaskStarted = QtCore.Signal(str)  # task name
    sigTaskStateChanged = QtCore.Signal(str, str)  # task name, task state
    sigTaskFinished = QtCore.Signal(str, object, bool)  # task name, result, success flag
    _sigStartTask = QtCore.Signal(str, dict)  # task name, args, kwargs

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._thread_lock = Mutex()
        self._running_tasks = dict()
        self._configured_task_types = dict()
        self._consecutive_activation = False  # Flag indicating consecutive activations

    def on_activate(self) -> None:
        """Initialise task runner."""
        self._running_tasks = dict()
        self._configured_task_types = dict()
        for name, task_cfg in self._module_task_configs.items():
            if name in self._configured_task_types:
                raise KeyError(f'Duplicate task name "{name}" encountered in config')
            module, cls = task_cfg['module.Class'].rsplit('.', 1)
            task = import_module_script(module, cls, reload=self._consecutive_activation)
            if not issubclass(task, ModuleTask):
                raise TypeError('Configured task is not a ModuleTask (sub)class')
            self._configured_task_types[name] = task
        self._sigStartTask.connect(self._run_task, QtCore.Qt.QueuedConnection)
        self._consecutive_activation = True

    def on_deactivate(self) -> None:
        """Shut down task runner."""
        self._sigStartTask.disconnect()
        for task in self._running_tasks.values():
            task.interrupt
        self._configured_task_types = dict()

    @property
    def running_tasks(self) -> List[str]:
        with self._thread_lock:
            return list(self._running_tasks)

    @property
    def task_states(self) -> Dict[str, str]:
        with self._thread_lock:
            states = dict()
            for task_name in self._configured_task_types:
                try:
                    states[task_name] = self._running_tasks[task_name].state
                except KeyError:
                    states[task_name] = 'stopped'
            return states

    @property
    def configured_task_types(self) -> Dict[str, Type[ModuleTask]]:
        return self._configured_task_types.copy()

    def run_task(self, name: str, arguments: Mapping[str, Any]) -> None:
        with self._thread_lock:
            self._sigStartTask.emit(name, dict(arguments))

    def interrupt_task(self, name: str) -> None:
        with self._thread_lock:
            task = self._running_tasks.get(name, None)
            if task is None:
                raise RuntimeError(f'No ModuleTask with name "{name}" running')
            task.interrupt()

    @QtCore.Slot(str, dict)
    def _run_task(self, name: str, arguments: Mapping[str, Any]) -> None:
        with self._thread_lock:
            task = self.__init_task(name)
            self.__set_task_arguments(task, arguments)
            self.__activate_connect_task_modules(name, task)
            self.__move_task_into_thread(name, task)
            self.__connect_task_signals(name, task)
            self.__start_task(name, task)
            self.sigTaskStarted.emit(name)

    def _task_finished_callback(self, name: str) -> None:
        """Called every time a task finishes."""
        with self._thread_lock:
            task = self._running_tasks.get(name, None)
            if task is not None:
                task.sigFinished.disconnect()
                task.sigStateChanged.disconnect()
                task.disconnect_modules()
                thread_manager = self._qudi_main.thread_manager
                thread_manager.quit_thread(task.thread())
                thread_manager.join_thread(task.thread())

    def _thread_finished_callback(self, name: str) -> None:
        with self._thread_lock:
            task = self._running_tasks.pop(name, None)
            if task is not None:
                self.sigTaskFinished.emit(name, task.result, task.success)

    def _task_state_changed_callback(self, state: str, name: str) -> None:
        self.sigTaskStateChanged.emit(name, state)

    def __init_task(self, name: str) -> ModuleTask:
        """Create a ModuleTask instance."""
        try:
            if name in self._running_tasks:
                raise RuntimeError(f'ModuleTask "{name}" is already initialized')
            return self._configured_task_types[name]()
        except:
            self.log.exception(f'Exception during initialization of ModuleTask "{name}":')
            raise

    def __set_task_arguments(self, task: ModuleTask, arguments: Mapping[str, Any]) -> None:
        """Set arguments for ModuleTask instance."""
        try:
            if not (isinstance(arguments, Mapping) and all(isinstance(a, str) for a in arguments)):
                raise TypeError('ModuleTask kwargs must be mapping with str type keys')
            task.kwargs = arguments
        except:
            self.log.exception(f'Exception during setting of arguments for ModuleTask:')
            raise

    def __activate_connect_task_modules(self, name: str, task: ModuleTask) -> None:
        """Activate and connect all configured module connectors for ModuleTask."""
        try:
            module_manager = self._qudi_main.module_manager
            connect_targets = dict()
            for conn_name, module_name in self._module_task_configs[name]['connect'].items():
                module = module_manager[module_name]
                module.activate()
                connect_targets[conn_name] = module.instance
            task.connect_modules(connect_targets)
        except:
            self.log.exception(f'Exception during modules connection for ModuleTask "{name}":')
            task.disconnect_modules()
            raise

    def __move_task_into_thread(self, name: str, task: ModuleTask) -> None:
        """Create a new QThread via qudi thread manager and move ModuleTask instance into it."""
        try:
            thread = self._qudi_main.thread_manager.get_new_thread(name=f'ModuleTask-{name}')
            if thread is None:
                raise RuntimeError(f'Unable to create QThread with name "ModuleTask-{name}"')
        except RuntimeError:
            self.log.exception('Exception during thread creation:')
            raise
        task.moveToThread(thread)
        thread.started.connect(task.run, QtCore.Qt.QueuedConnection)
        thread.finished.connect(partial(self._thread_finished_callback, name=name))

    def __connect_task_signals(self, name: str, task: ModuleTask) -> None:
        task.sigFinished.connect(partial(self._task_finished_callback, name=name),
                                 QtCore.Qt.QueuedConnection)
        task.sigStateChanged.connect(partial(self._task_state_changed_callback, name=name),
                                     QtCore.Qt.QueuedConnection)

    def __start_task(self, name: str, task: ModuleTask) -> None:
        self._running_tasks[name] = task
        task.thread().start()



# -*- coding: utf-8 -*-
"""
This file contains Qudi methods for data filtering.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('scan_blink_correction',)

import numpy as np
from scipy.ndimage import minimum_filter1d, maximum_filter1d
import logging

_logger = logging.getLogger(__name__)


def scan_blink_correction(image, axis=1):
    """
    Filter out impulsive noise from a 2D array along a single axis using an opening filter.

    The opening filter applies a sequence of two filters: first a min-filter and then a max-filter.
    This technique is effective at removing single-pixel brightness spikes along the specified axis,
    but it may make the image appear more "blocky" or less smooth.

    Parameters
    ----------
    image : numpy.ndarray
        A 2D numpy array to be filtered (e.g., image data).
    axis : int
        The axis along which to apply the 1D filter.

    Returns
    -------
    numpy.ndarray
        The filtered image, with the same dimensions as the input image.

    Notes
    -----
    Ensure that the image features of interest are larger than the impulsive noise spikes
    to achieve effective noise reduction without loss of significant image detail.

    """

    if not isinstance(image, np.ndarray):
        _logger.error('Image must be 2D numpy array.')
        return image
    if image.ndim != 2:
        _logger.error('Image must be 2D numpy array.')
        return image
    if axis != 0 and axis != 1:
        _logger.error('Optional axis parameter must be either 0 or 1.')
        return image

    # Calculate median value of the image. This value is used for padding image boundaries during
    # filtering.
    median = np.median(image)
    # Apply a minimum filter along the chosen axis.
    filt_img = minimum_filter1d(image, size=2, axis=axis, mode='constant', cval=median)
    # Apply a maximum filter along the chosen axis. Flip the previous filter result to avoid
    # translation of image features.
    filt_img = maximum_filter1d(
        np.flip(filt_img, axis), size=2, axis=axis, mode='constant', cval=median)
    # Flip back the image to obtain original orientation and return result.
    return np.flip(filt_img, axis)



# -*- coding: utf-8 -*-

"""
This file contains a custom Colorbar Widget to be used with pyqtgraph.ImageItem or qudi
ScanImageItem.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ColorBarWidget']

from enum import IntEnum
from pyqtgraph import mkPen, mkBrush, PlotWidget, BarGraphItem
from PySide2 import QtCore, QtGui, QtWidgets
from qudi.util.widgets.scientific_spinbox import ScienDSpinBox
from qudi.util.colordefs import ColorScaleInferno


class ColorBarItem(BarGraphItem):
    def __init__(self, parent=None, limits=(0, 1), cmap=None, pen=None):
        limits = (float(min(limits)), float(max(limits)))
        cmap = ColorScaleInferno().colormap if cmap is None else cmap
        pen = mkPen(QtGui.QPen(QtCore.Qt.PenStyle.NoPen)) if pen is None else mkPen(pen)
        grad = QtGui.QLinearGradient(0, 0, 0, 1)
        grad.setCoordinateMode(QtGui.QGradient.ObjectMode)
        for stop, color in zip(*cmap.getStops('byte')):
            grad.setColorAt(stop, QtGui.QColor(*color))
        brush = mkBrush(QtGui.QBrush(grad))
        height = abs(limits[1] - limits[0])
        super().__init__(parent=parent,
                         x=[0],
                         y=[limits[0] + height / 2],
                         height=[height],
                         width=1.5,
                         brush=brush,
                         pen=pen)

    def set_limits(self, min_val, max_val):
        if max_val < min_val:
            min_val, max_val = max_val, min_val
        height = abs(max_val - min_val)
        self.setOpts(y=[min_val + height / 2], height=[height])


class ColorBarWidget(QtWidgets.QWidget):
    """ A widget containing a controllable colorbar for color-coded plots.
    """

    class ColorBarMode(IntEnum):
        ABSOLUTE = 0
        PERCENTILE = 1

    sigLimitsChanged = QtCore.Signal(tuple)  # (min_val, max_val)
    sigPercentilesChanged = QtCore.Signal(tuple)  # (low_percentile, high_percentile)
    sigModeChanged = QtCore.Signal(object)

    def __init__(self, *args, unit=None, label=None, absolute_range=None, percentile_range=None,
                 mode=ColorBarMode.PERCENTILE, **kwargs):
        super().__init__(*args, **kwargs)

        self.min_spinbox = ScienDSpinBox()
        self.min_spinbox.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                       QtWidgets.QSizePolicy.Fixed)
        self.min_spinbox.setAlignment(QtCore.Qt.AlignRight)
        self.min_spinbox.setMinimumWidth(75)
        self.min_spinbox.setValue(0)
        self.max_spinbox = ScienDSpinBox()
        self.max_spinbox.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                       QtWidgets.QSizePolicy.Fixed)
        self.max_spinbox.setAlignment(QtCore.Qt.AlignRight)
        self.min_spinbox.setMinimumWidth(75)
        self.max_spinbox.setValue(1)
        self.low_percentile_spinbox = ScienDSpinBox()
        self.low_percentile_spinbox.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                                  QtWidgets.QSizePolicy.Fixed)
        self.low_percentile_spinbox.setAlignment(QtCore.Qt.AlignRight)
        self.low_percentile_spinbox.setMinimumWidth(75)
        self.low_percentile_spinbox.setSuffix('%')
        self.low_percentile_spinbox.setMinimalStep('0.01')
        self.low_percentile_spinbox.setValue(0)
        self.high_percentile_spinbox = ScienDSpinBox()
        self.high_percentile_spinbox.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                                   QtWidgets.QSizePolicy.Fixed)
        self.high_percentile_spinbox.setAlignment(QtCore.Qt.AlignRight)
        self.high_percentile_spinbox.setMinimumWidth(75)
        self.high_percentile_spinbox.setSuffix('%')
        self.high_percentile_spinbox.setMinimalStep('0.01')
        self.high_percentile_spinbox.setValue(100)
        if unit is not None:
            self.max_spinbox.setSuffix(unit)
            self.min_spinbox.setSuffix(unit)
        if absolute_range is not None:
            self.min_spinbox.setRange(*absolute_range)
            self.max_spinbox.setRange(*absolute_range)
        if percentile_range is not None:
            min_percentile = percentile_range[0] if 0 <= percentile_range[0] <= 100 else 0
            max_percentile = percentile_range[1] if 0 <= percentile_range[1] <= 100 else 0
            self.low_percentile_spinbox.setRange(min_percentile, max_percentile)
            self.high_percentile_spinbox.setRange(min_percentile, max_percentile)
        else:
            self.low_percentile_spinbox.setRange(0, 100)
            self.high_percentile_spinbox.setRange(0, 100)

        grad = QtGui.QLinearGradient(0, 0, 0, 1)
        grad.setCoordinateMode(QtGui.QGradient.ObjectMode)
        for stop, color in zip(*ColorScaleInferno().colormap.getStops('byte')):
            grad.setColorAt(stop, QtGui.QColor(*color))
        self._cb_brush = mkBrush(QtGui.QBrush(grad))

        self.colorbar = ColorBarItem()
        self.cb_plot_widget = PlotWidget()
        self.cb_plot_widget.hideButtons()
        self.cb_plot_widget.setMinimumWidth(75)
        self.cb_plot_widget.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self.cb_plot_widget.addItem(self.colorbar)
        self.cb_plot_widget.hideAxis('bottom')
        self.cb_plot_widget.setLabel('left', text=label, units=unit)
        self.cb_plot_widget.setMouseEnabled(x=False, y=False)
        self.cb_plot_widget.disableAutoRange()
        self.cb_plot_widget.setYRange(0, 1)
        self.cb_plot_widget.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        self.absolute_radioButton = QtWidgets.QRadioButton('Absolute')
        self.absolute_radioButton.setAutoExclusive(True)
        self.percentile_radioButton = QtWidgets.QRadioButton('Percentile')
        self.percentile_radioButton.setAutoExclusive(True)

        main_layout = QtWidgets.QVBoxLayout()
        main_layout.addWidget(self.max_spinbox)
        main_layout.addWidget(self.high_percentile_spinbox)
        main_layout.addWidget(self.cb_plot_widget)
        main_layout.addWidget(self.low_percentile_spinbox)
        main_layout.addWidget(self.min_spinbox)
        main_layout.addWidget(self.absolute_radioButton)
        main_layout.addWidget(self.percentile_radioButton)

        if mode == self.ColorBarMode.ABSOLUTE:
            self.absolute_radioButton.setChecked(True)
        else:
            self.percentile_radioButton.setChecked(True)

        # main_layout.setSpacing(0)
        main_layout.setContentsMargins(1, 1, 1, 1)
        self.setLayout(main_layout)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        # Connect signals
        self.min_spinbox.valueChanged.connect(self._absolute_value_changed)
        self.max_spinbox.valueChanged.connect(self._absolute_value_changed)
        self.low_percentile_spinbox.valueChanged.connect(self._percentile_value_changed)
        self.high_percentile_spinbox.valueChanged.connect(self._percentile_value_changed)
        self.percentile_radioButton.toggled.connect(self._mode_changed)
        self.absolute_radioButton.toggled.connect(self._mode_changed)
        return

    @property
    def mode(self):
        if self.absolute_radioButton.isChecked():
            return self.ColorBarMode.ABSOLUTE
        return self.ColorBarMode.PERCENTILE

    @property
    def limits(self):
        return self.min_spinbox.value(), self.max_spinbox.value()

    @property
    def percentiles(self):
        return self.low_percentile_spinbox.value(), self.high_percentile_spinbox.value()

    # FIXME: Remove?
    def sizeHint(self):
        return QtCore.QSize(90, 100)

    def set_label(self, text, unit=None):
        if unit is not None:
            self.max_spinbox.setSuffix(unit)
            self.min_spinbox.setSuffix(unit)
        return self.cb_plot_widget.setLabel('left', text=text, units=unit)

    def set_colormap(self, cmap=None):
        return self.colorbar.set_cmap(cmap=cmap)

    def set_pen(self, pen=None):
        return self.colorbar.set_pen(pen)

    @QtCore.Slot(float, float, float, float)
    def set_limits(self, min_value, max_value, low_percentile=None, high_percentile=None):
        # Check and set percentile values in spinboxes
        if (low_percentile is None) != (high_percentile is None):
            raise ValueError('If percentile ranges should be changed, you must specify both low '
                             'and high percentile values.')
        elif low_percentile is not None:
            self.low_percentile_spinbox.blockSignals(True)
            self.high_percentile_spinbox.blockSignals(True)
            self.low_percentile_spinbox.setValue(low_percentile)
            self.high_percentile_spinbox.setValue(high_percentile)
            self.low_percentile_spinbox.blockSignals(False)
            self.high_percentile_spinbox.blockSignals(False)

        # Set absolute values in spinboxes and update colorbar
        self.min_spinbox.blockSignals(True)
        self.max_spinbox.blockSignals(True)
        self.min_spinbox.setValue(min_value)
        self.max_spinbox.setValue(max_value)
        min_val = self.min_spinbox.value()
        max_val = self.max_spinbox.value()
        self.colorbar.set_limits(min_val, max_val)
        self.cb_plot_widget.setYRange(min_val, max_val)

        self.min_spinbox.blockSignals(False)
        self.max_spinbox.blockSignals(False)

    @QtCore.Slot(object)
    def set_mode(self, mode):
        mode = self.ColorBarMode(mode)
        if mode is self.ColorBarMode.ABSOLUTE:
            self.absolute_radioButton.setChecked(True)
        else:
            self.percentile_radioButton.setChecked(True)
        return

    @QtCore.Slot()
    def _absolute_value_changed(self):
        min_val = self.min_spinbox.value()
        max_val = self.max_spinbox.value()
        self.colorbar.set_limits(min_val, max_val)
        self.cb_plot_widget.setYRange(min_val, max_val)
        if not self.absolute_radioButton.isChecked():
            self.absolute_radioButton.setChecked(True)
            self.sigModeChanged.emit(self.ColorBarMode.ABSOLUTE)
        self.sigLimitsChanged.emit((min_val, max_val))
        return

    @QtCore.Slot()
    def _percentile_value_changed(self):
        if not self.percentile_radioButton.isChecked():
            self.percentile_radioButton.setChecked(True)
            self.sigModeChanged.emit(self.ColorBarMode.PERCENTILE)
        self.sigPercentilesChanged.emit((self.low_percentile_spinbox.value(),
                                         self.high_percentile_spinbox.value()))
        return

    @QtCore.Slot()
    def _mode_changed(self):
        if self.absolute_radioButton.isChecked():
            self.sigModeChanged.emit(self.ColorBarMode.ABSOLUTE)
        else:
            self.sigModeChanged.emit(self.ColorBarMode.PERCENTILE)
        return



# -*- coding: utf-8 -*-
"""
This module contains a QWidgets.QDialog subclass representing an "about qudi" dialog.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from PySide2 import QtCore, QtWidgets


class AboutQudiDialog(QtWidgets.QDialog):
    """
    QWidgets.QDialog subclass representing an "about qudi" dialog.
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.setWindowFlags(QtCore.Qt.WindowTitleHint | QtCore.Qt.WindowCloseButtonHint)

        buttonbox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        buttonbox.setOrientation(QtCore.Qt.Horizontal)
        self.ok_button = buttonbox.button(buttonbox.Ok)
        self.ok_button.clicked.connect(self.accept)

        self.header_label = QtWidgets.QLabel('qudi')
        self.header_label.setObjectName('headerLabel')
        font = self.header_label.font()
        font.setBold(True)
        font.setPointSize(20)
        self.header_label.setFont(font)
        self.version_label = QtWidgets.QLabel('Version number goes here...')
        self.version_label.setObjectName('versionLabel')
        self.version_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse
                                                   | QtCore.Qt.TextBrowserInteraction)
        self.version_label.setOpenExternalLinks(True)

        self.about_label = QtWidgets.QLabel('<html><head/><body><p>Qudi is a suite of tools for '
                                            'operating multi-instrument and multi-computer '
                                            'laboratory experiments. Originally built around a '
                                            'confocal fluorescence microscope experiments, it has '
                                            'grown to be a generally applicaple framework for '
                                            'controlling experiments.</p></body></html>')
        self.about_label.setWordWrap(True)
        self.about_label.setObjectName('aboutLabel')
        self.about_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse
                                                 | QtCore.Qt.TextBrowserInteraction)
        self.about_label.setOpenExternalLinks(True)

        self.credits_label = QtWidgets.QLabel(
            '<html><head/><body><p><span style=" text-decoration: underline;">Qudi was originally '
            'developed by the Institute for Quantum Optics at Ulm University. </span></p><p>'
            'You can find a full list of contributors at the '
            '<a href="https://github.com/Ulm-IQO/qudi-core/graphs/contributors">'
            '<span style=" text-decoration: underline; color:#00ffff;">qudi-core GitHub project '
            'page</span></a>.</p></body></html>'
        )
        self.credits_label.setWordWrap(True)
        self.credits_label.setObjectName('creditsLabel')
        self.credits_label.setTextInteractionFlags(
            QtCore.Qt.TextSelectableByMouse | QtCore.Qt.TextBrowserInteraction
        )
        self.credits_label.setOpenExternalLinks(True)

        self.license_label = QtWidgets.QLabel(
            '<html>\n<head/>\n<body>\n<p><span style=" font-family:"monospace";">\n'
            'Qudi is free software: you can redistribute it and/or modify it under the terms '
            '<br/>\nof the GNU Lesser General Public License as published by the Free Software '
            'Foundation,<br/>\neither version 3 of the License, or (at your option) any later '
            'version.\n</span></p>\n<p><span style=" font-family:"monospace";">\nQudi is '
            'distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;<br/>\n'
            'without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR '
            'PURPOSE.<br/>\nSee the GNU Lesser General Public License for more details.<br/>\n'
            '</span>\n</p><p><span style=" font-family:"monospace";">\nYou should have received a '
            'copy of the GNU Lesser General Public License along with qudi.<br/>\nIf not, see\n'
            '</span>\n<a href="https://www.gnu.org/licenses/"><span style=" text-decoration: '
            'underline; color:#00ffff;">https://www.gnu.org/licenses/</span></a>.<br/>\n</p>\n'
            '<p><span style=" text-decoration: underline;">\nQudi is derived in parts from ACQ4, '
            'so here is its license:</span></p>\n<p><span style=" font-family:"monospace";">\n'
            'Permission is hereby granted, free of charge, to any person obtaining a copy <br/>\nof'
            ' this software and associated documentation files (the &quot;Software&quot;), to deal '
            '<br/>\nin the Software without restriction, including without limitation the rights '
            '<br/>\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell '
            '<br/>\ncopies of the Software, and to permit persons to whom the Software is <br/>\n'
            'furnished to do so, subject to the following conditions: <br/>\n<br/>\nThe above '
            'copyright notice and this permission notice shall be included in all <br/>\ncopies or '
            'substantial portions of the Software. <br/>\n<br/>\nTHE SOFTWARE IS PROVIDED &quot;AS '
            'IS&quot;, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR <br/>\nIMPLIED, INCLUDING BUT NOT '
            'LIMITED TO THE WARRANTIES OF MERCHANTABILITY, <br/>\nFITNESS FOR A PARTICULAR PURPOSE '
            'AND NONINFRINGEMENT. IN NO EVENT SHALL THE <br/>\nAUTHORS OR COPYRIGHT HOLDERS BE '
            'LIABLE FOR ANY CLAIM, DAMAGES OR OTHER <br/>\nLIABILITY, WHETHER IN AN ACTION OF '
            'CONTRACT, TORT OR OTHERWISE, ARISING FROM, <br/>\nOUT OF OR IN CONNECTION WITH THE '
            'SOFTWARE OR THE USE OR OTHER DEALINGS IN THE <br/>\nSOFTWARE.<br/>\n</span></p>\n'
            '<p><span style=" font-family:"monospace"; text-decoration: underline;">\n<br/>\n<br/>'
            '\n</span></p>\n</body>\n</html>'
        )
        self.license_label.setWordWrap(True)
        self.license_label.setOpenExternalLinks(True)
        self.license_label.setObjectName('licenseLabel')
        self.license_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse
                                                   | QtCore.Qt.TextBrowserInteraction)
        self.license_label.setOpenExternalLinks(True)

        self.citation_label = QtWidgets.QLabel(
            '<html><head/><body><p>'
            'If you are publishing scientific results, mentioning Qudi in your methods decscription <br/>\n'
            'is the least you can do as good scientific practice. You should cite our paper <br/>\n<br/>\n'
            'Qudi: A modular python suite for experiment control and data processing.<br/>\n'
            '</span>\n<a href="https://doi.org/10.1016/j.softx.2017.02.001"><span style=" text-decoration: '
            'underline; color:#00ffff;">https://doi.org/10.1016/j.softx.2017.02.001</span></a><br/>\n</p>\n'
            '@article{BINDER201785,<br/>\n'
            'title = {Qudi: A modular python suite for experiment control and data processing},<br/>\n'
            'journal = {SoftwareX},<br/>\n'
            'volume = {6},<br/>\n'
            'pages = {85-90},<br/>\n'
            'year = {2017},<br/>\n'
            'issn = {2352-7110},<br/>\n'
            'doi = {https://doi.org/10.1016/j.softx.2017.02.001},<br/>\n'
            'url = {https://www.sciencedirect.com/science/article/pii/S2352711017300055},<br/>\n'
            'author = {Jan M. Binder and Alexander Stark and Nikolas Tomek and Jochen Scheuer and Florian Frank and Kay D. Jahnke and Christoph Müller and Simon Schmitt and Mathias H. Metsch and Thomas Unden and Tobias Gehring and Alexander Huck and Ulrik L. Andersen and Lachlan J. Rogers and Fedor Jelezko},<br/>\n'
            'keywords = {Python 3, Qt, Experiment control, Automation, Measurement software, Framework, Modular},<br/>\n'
            'abstract = {Qudi is a general, modular, multi-operating system suite written in Python 3 for controlling laboratory experiments. It provides a structured environment by separating functionality into hardware abstraction, experiment logic and user interface layers. The core feature set comprises a graphical user interface, live data visualization, distributed execution over networks, rapid prototyping via Jupyter notebooks, configuration management, and data recording. Currently, the included modules are focused on confocal microscopy, quantum optics and quantum information experiments, but an expansion into other fields is possible and encouraged.}<br/>\n'
            '}<br/>\n'
            '</p></body></html>'
        )
        self.citation_label.setWordWrap(True)
        self.citation_label.setObjectName('creditsLabel')
        self.citation_label.setTextInteractionFlags(
            QtCore.Qt.TextSelectableByMouse | QtCore.Qt.TextBrowserInteraction
        )
        self.citation_label.setOpenExternalLinks(True)

        about_scroll_widget = QtWidgets.QScrollArea()
        about_scroll_widget.setWidgetResizable(True)
        about_scroll_widget.setWidget(self.about_label)
        about_scroll_widget.setObjectName('aboutScrollArea')
        credits_scroll_widget = QtWidgets.QScrollArea()
        credits_scroll_widget.setWidgetResizable(True)
        credits_scroll_widget.setWidget(self.credits_label)
        credits_scroll_widget.setObjectName('creditsScrollArea')
        license_scroll_widget = QtWidgets.QScrollArea()
        license_scroll_widget.setWidgetResizable(True)
        license_scroll_widget.setWidget(self.license_label)
        license_scroll_widget.setObjectName('licenseScrollArea')
        citation_scroll_widget = QtWidgets.QScrollArea()
        citation_scroll_widget.setWidgetResizable(True)
        citation_scroll_widget.setWidget(self.citation_label)
        citation_scroll_widget.setObjectName('citationScrollArea')

        self.tab_widget = QtWidgets.QTabWidget()
        self.tab_widget.setObjectName('tabWidget')
        self.tab_widget.addTab(about_scroll_widget, 'About')
        self.tab_widget.addTab(credits_scroll_widget, 'Credits')
        self.tab_widget.addTab(license_scroll_widget, 'License')
        self.tab_widget.addTab(citation_scroll_widget, 'Cite Qudi')

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.header_label)
        layout.addWidget(self.version_label)
        layout.addWidget(self.tab_widget)
        layout.addWidget(buttonbox)

        self.setLayout(layout)
        self.about_label.setFocus()
        return
   



# -*- coding: utf-8 -*-
"""
This file contains the QMainWindow class for the task GUI.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import os
from typing import Any, Mapping, Dict, Type, Callable
from PySide2 import QtCore, QtGui, QtWidgets

from qudi.util.paths import get_artwork_dir
from qudi.core.scripting.moduletask import ModuleTask

from .taskwidget import TaskWidget


class TaskMainWindow(QtWidgets.QMainWindow):
    """
    Main Window definition for the task GUI.
    """

    sigStartTask = QtCore.Signal(str, dict)  # task name, call parameters
    sigInterruptTask = QtCore.Signal(str)  # task name
    sigClosed = QtCore.Signal()

    def __init__(self, *args, tasks: Mapping[str, Type[ModuleTask]], **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle('qudi: Taskrunner')

        # Create actions
        icon_path = os.path.join(get_artwork_dir(), 'icons')
        self.action_quit = QtWidgets.QAction()
        self.action_quit.setIcon(QtGui.QIcon(os.path.join(icon_path, 'application-exit')))
        self.action_quit.setText('Close')
        self.action_quit.setToolTip('Close')
        self.action_quit.triggered.connect(self.close)

        # Create menu bar
        self.menubar = QtWidgets.QMenuBar()
        menu = QtWidgets.QMenu('File')
        menu.addAction(self.action_quit)
        self.menubar.addMenu(menu)
        self.setMenuBar(self.menubar)

        # Create central container widget for ModuleTask widgets
        # self.scroll_area = QtWidgets.QScrollArea()
        self.task_widgets = dict()
        self.tasks_layout = QtWidgets.QVBoxLayout()
        widget = QtWidgets.QWidget()
        widget.setLayout(self.tasks_layout)
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidget(widget)
        scroll_area.setWidgetResizable(True)
        self.setCentralWidget(scroll_area)

        self._initialize_task_widgets(tasks)

        # # Create toolbar
        # self.toolbar = QtWidgets.QToolBar()
        # self.toolbar.setOrientation(QtCore.Qt.Horizontal)
        # self.toolbar.addAction(self.action_start_task)
        # self.toolbar.addAction(self.action_pause_task)
        # self.toolbar.addAction(self.action_stop_task)
        # self.addToolBar(self.toolbar)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        super().closeEvent(event)
        self.sigClosed.emit()

    @QtCore.Slot(str)
    def task_started(self, name: str) -> None:
        self.task_widgets[name].task_started()

    @QtCore.Slot(str, str)
    def task_state_changed(self, name: str, state: str) -> None:
        self.task_widgets[name].task_state_changed(state)

    @QtCore.Slot(str, object, bool)
    def task_finished(self, name: str, result: Any, success: bool) -> None:
        self.task_widgets[name].task_finished(result, success)

    def _initialize_task_widgets(self, tasks: Mapping[str, Type[ModuleTask]]) -> None:
        for ii, (task_name, task_type) in enumerate(tasks.items()):
            groupbox = QtWidgets.QGroupBox(task_name)
            font = groupbox.font()
            font.setBold(True)
            font.setPointSize(font.pointSize() + 2)
            groupbox.setFont(font)
            widget = TaskWidget(task_type=task_type)
            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(widget)
            groupbox.setLayout(layout)
            widget.sigStartTask.connect(self._get_start_task_callback(task_name))
            widget.sigInterruptTask.connect(self._get_interrupt_task_callback(task_name))
            self.tasks_layout.addWidget(groupbox)
            self.task_widgets[task_name] = widget

    def _clear_task_widgets(self) -> None:
        """Helper method to disconnect and delete all TaskWidgets and remove them from layout.
        """
        for widget in reversed(self.task_widgets):
            groupbox = widget.parent()
            widget.sigStartTask.disconnect()
            widget.sigInterruptTask.disconnect()
            self.tasks_layout.removeWidget(groupbox)
            groupbox.setParent(None)
            groupbox.deleteLater()
        self.task_widgets = dict()

    def _get_start_task_callback(self, task_name: str) -> Callable[[Dict[str, Any]], None]:

        def callback(parameters: Dict[str, Any]) -> None:
            self.sigStartTask.emit(task_name, parameters)

        return callback

    def _get_interrupt_task_callback(self, task_name: str) -> Callable[[], None]:

        def callback() -> None:
            self.sigInterruptTask.emit(task_name)

        return callback



# -*- coding: utf-8 -*-

"""
JSON (draft v7) validator for qudi YAML configurations that also fills in default values.
The corresponding JSON schema is defined in ".__schema.py".

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ValidationError', 'validate_config', 'validate_local_module_config',
           'validate_remote_module_config', 'validate_module_name']

import re
from typing import Mapping, Any
from jsonschema import ValidationError
from jsonschema import validators as __validators
from jsonschema import Draft7Validator as __BaseValidator

from .schema import config_schema, remote_module_config_schema, local_module_config_schema


def __set_defaults(validator, properties, instance, schema):
    # Only insert default values of current schema into instance if validation passses
    try:
        __BaseValidator(schema).validate(instance)
    except ValidationError:
        pass
    else:
        for property, subschema in properties.items():
            if 'default' in subschema:
                try:
                    instance.setdefault(property, subschema['default'])
                except AttributeError:
                    pass

    for error in __BaseValidator.VALIDATORS['properties'](validator, properties, instance, schema):
        yield error


def __is_iterable(checker, instance):
    return (__BaseValidator.TYPE_CHECKER.is_type(instance, "array") or
            isinstance(instance, (set, frozenset, tuple)))


# Add custom JSON schema (draft v7) validator that accepts all Python builtin sequences as "array"
# type
DefaultInsertionValidator = __validators.extend(
    validator=__BaseValidator,
    validators={'properties': __set_defaults},
    type_checker=__BaseValidator.TYPE_CHECKER.redefine("array", __is_iterable)
)


def validate_config(config: Mapping[str, Any]) -> None:
    """JSON schema (draft v7) validator for qudi configuration.
    Raises jsonschema.ValidationError if invalid.
    """
    DefaultInsertionValidator(config_schema()).validate(config)


def validate_local_module_config(config: Mapping[str, Any]) -> None:
    """JSON schema (draft v7) validator for single qudi local module configuration.
    Raises jsonschema.ValidationError if invalid.
    """
    DefaultInsertionValidator(local_module_config_schema()).validate(config)


def validate_remote_module_config(config: Mapping[str, Any]) -> None:
    """JSON schema (draft v7) validator for single qudi remote module configuration.
    Raises jsonschema.ValidationError if invalid.
    """
    DefaultInsertionValidator(remote_module_config_schema()).validate(config)


def validate_module_name(name: str) -> None:
    """Regular expression validator for qudi module names.
    Raises jsonschema.ValidationError if invalid.

    WARNING: The jsonschema.ValidationError raised does not contain any JSON schema information.
    """
    if re.match(r'^[a-zA-Z_]+[a-zA-Z0-9_]*$', name) is None:
        raise ValidationError(
            'Module names must only contain word characters [a-zA-Z0-9_] and not start on a number.'
        )



# -*- coding: utf-8 -*-
"""
This file contains a custom QAbstractTableModel object providing text data for all logged records.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('LogRecordsTableModel',)

import traceback
from datetime import datetime
from PySide2 import QtCore, QtGui
from qudi.util.mutex import Mutex


class LogRecordsTableModel(QtCore.QAbstractTableModel):
    """This is a Qt model that represents textual information about all logged records.
    Can be displayed with a QTableView for example.
    """

    _color_map = {'debug'   : QtGui.QColor('#77F'),
                  'info'    : QtGui.QColor('#1F1'),
                  'warning' : QtGui.QColor('#F90'),
                  'error'   : QtGui.QColor('#F11'),
                  'critical': QtGui.QColor('#FF00FF'),
                  }
    _fallback_color = QtGui.QColor('#FFF')
    _header = ('Time', 'Level', 'Source', 'Message')

    def __init__(self, *args, max_records=10000, **kwargs):
        super().__init__(*args, **kwargs)

        self._thread_lock = Mutex()
        self._max_records = max(int(max_records), 1)
        self._records = list()
        self._begin = 0
        self._end = 0
        self._fill_count = 0

    def rowCount(self, parent=None):
        """Returns the number of log records stored in the model.

        Returns
        -------
        int
            Number of log records stored.
        """
        return self._fill_count

    def columnCount(self, parent=None):
        """Returns the number of columns each log record has.

        Returns
        -------
        int
            Number of columns for each log record.
        """
        return len(self._header)

    def flags(self, index):
        """Determines what can be done with log record cells in the table view.

        Parameters
        ----------
        index : QModelIndex
            Cell for which the flags are requested.

        Returns
        -------
        Qt.ItemFlags
            Actions allowed for this cell.
        """
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable

    def data(self, index, role):
        """Get data from model for a given cell. Data can have a role that affects display.

        Parameters
        ----------
        index : QModelIndex
            Cell for which data is requested.
        role : ItemDataRole
            Role for which data is requested.

        Returns
        -------
        QVariant
            Data for given cell and role.
        """
        if index.isValid():
            record = self._records[(self._begin + index.row()) % self._max_records]
            if role == QtCore.Qt.TextColorRole:
                return self._color_map.get(record[1], self._fallback_color)
            if role in (QtCore.Qt.DisplayRole, QtCore.Qt.ToolTipRole, QtCore.Qt.EditRole):
                return record[index.column()]

    def headerData(self, section, orientation, role=None):
        """Data for the table view headers.

        Parameters
        ----------
        section : int
            Number of the column to get header data for.
        orientation : Qt.Orientation
            Orientation of header (horizontal or vertical).
        role : ItemDataRole
            Role for which to get data.

        Returns
        -------
        QVariant
            Header data for given column and role.
        """
        if (role is None or role == QtCore.Qt.DisplayRole) and orientation == QtCore.Qt.Horizontal:
            try:
                return self._header[section]
            except IndexError:
                pass
        return

    @QtCore.Slot(object)
    def add_record(self, data):
        """Add a single log entry to the end of the table model.

        Parameters
        ----------
        data : logging.LogRecord
            Log record as returned from logging module.

        Returns
        -------
        bool
            True if adding entry succeeded, False otherwise.
        """
        with self._thread_lock:
            if self._fill_count < self._max_records:
                self.beginInsertRows(QtCore.QModelIndex(), self._end, self._end)
                self._records.append(self._format_log_record(data))
                self._fill_count += 1
                self._end = (self._end + 1) % self._max_records
                self.endInsertRows()
            else:
                row = self._max_records - 1
                self.beginRemoveRows(QtCore.QModelIndex(), 0, 0)
                self._begin = (self._begin + 1) % self._max_records
                self._fill_count -= 1
                self.endRemoveRows()

                self.beginInsertRows(QtCore.QModelIndex(), row, row)
                self._records[self._end] = self._format_log_record(data)
                self._end = (self._end + 1) % self._max_records
                self._fill_count += 1
                self.endInsertRows()

    @QtCore.Slot()
    def clear(self):
        with self._thread_lock:
            self.beginResetModel()
            self._begin = 0
            self._end = 0
            self._fill_count = 0
            self._records = list()
            self.endResetModel()

    @property
    def max_size(self):
        return self._max_records

    @staticmethod
    def _format_log_record(record):
        # Compose message to display
        message = record.getMessage()  # message if hasattr(record, 'message') else record.msg
        if record.exc_info is not None:
            message += f'\n\n{traceback.format_exception(*record.exc_info)[-1][:-1]}'
            tb = '\n'.join(traceback.format_exception(*record.exc_info)[:-1])
            if tb:
                message += f'\n{tb}'

        # Create human-readable timestamp
        timestamp = datetime.fromtimestamp(record.created).strftime('%Y-%m-%d %H:%M:%S')

        # return 4 element tuple (timestamp, level, name, message)
        # Avoid problems with Qt by eliminating NULL bytes in strings.
        return timestamp, record.levelname, record.name, message.replace('\0', '\\x00')



# -*- coding: utf-8 -*-

"""
This file contains a custom .ui file loader since the current (v5.14.1) Pyside2 implementation or
qtpy implementation do not fully allow promotion to a custom widget if the custom widget is not a
direct subclass of the base widget defined in the .ui file. For example you can subclass
QDoubleSpinBox and promote this to your custom class MyDoubleSpinBox but you can not properly
subclass QAbstractSpinBox and promote QDoubleSpinBox (even though QDoubleSpinBox inherits
QAbstractSpinBox).
Funny enough it works if you use Pyside2's ui-to-py-converter and run the generated python code.
This module provides a wrapper to do just that.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['loadUi']

import os
import re
import tempfile
import subprocess
from importlib.util import spec_from_loader, module_from_spec
from qudi.util.paths import get_artwork_dir

__ui_class_pattern = re.compile(r'class (Ui_.*?)\(')
__artwork_path_pattern = re.compile(r'>(.*?/artwork/.*?)</')


def loadUi(file_path, base_widget):
    """Compiles a given .ui-file at <file_path> into python code. This code will be executed and
    the generated class will be used to initialize the widget given in <base_widget>.
    Creates a temporary file in the systems tmp directory using the tempfile module.
    The original .ui file will remain untouched.

    WARNING: base_widget must be of the same class as the top-level widget in the .ui file.
             Compatible subclasses of the top-level widget in the .ui file will also work.

    Parameters
    ----------
    file_path : str
        The full path to the .ui-file to load.
    base_widget : object
        Instance of the base widget represented by the .ui-file.

    Returns
    -------
    None
    """
    # This step is a workaround because Qt Designer will only specify relative paths which is very
    # error prone if the user changes the cwd (e.g. os.chdir)
    converted = _convert_ui_to_absolute_paths(file_path)

    # Compile (converted) .ui-file into python code
    # If file needed conversion, write temporary file to be used in subprocess
    if converted is not None:
        fd, file_path = tempfile.mkstemp(text=True)
        try:
            with open(fd, mode='w', closefd=True) as tmp_file:
                tmp_file.write(converted)
        except:
            os.remove(file_path)
            raise
    try:
        result = subprocess.run(['pyside2-uic', file_path],
                                capture_output=True,
                                text=True,
                                check=True)
        compiled = result.stdout
    finally:
        if converted is not None:
            os.remove(file_path)

    # Find class name
    match = __ui_class_pattern.search(compiled)
    if match is None:
        raise RuntimeError('Failed to match regex for finding class name in generated python code.')
    class_name = match.groups()[0]
    # Workaround (again) because pyside2-uic forgot to include objects from PySide2 that can be
    # used by Qt Designer. So we inject import statements here just before the class declaration.
    insert = match.start()
    compiled = compiled[:insert] + 'from PySide2.QtCore import QLocale\n\n' + compiled[insert:]

    # Execute python code in order to obtain a module object from it
    spec = spec_from_loader('ui_module', loader=None)
    ui_module = module_from_spec(spec)
    exec(compiled, ui_module.__dict__)

    loader = getattr(ui_module, class_name, None)()
    if loader is None:
        raise RuntimeError('Unable to locate generated Ui_... class')
    loader.setupUi(base_widget)
    # Merge namespaces manually since this is not done by setupUi.
    to_merge = vars(loader)
    ignore = set(to_merge).intersection(set(base_widget.__dict__))  # Avoid namespace conflicts.
    for key in ignore:
        del to_merge[key]
    base_widget.__dict__.update(to_merge)


def _convert_ui_to_absolute_paths(file_path):
    """Converts the .ui file in order to change all relative path declarations containing the
    keyword "/artwork/" into absolute paths pointing to the qudi artwork data directory.

    Parameters
    ----------
    file_path : str
        The path to the .ui file to convert.

    Returns
    -------
    str | None
        Converted file content of the .ui file, None if conversion is not needed.
    """
    path_prefix = get_artwork_dir()
    with open(file_path, 'r') as file:
        ui_content = file.read()
    chunks = __artwork_path_pattern.split(ui_content)
    # Iterate over odd indices. Remember if changes were needed
    has_changed = False
    for ii in range(1, len(chunks), 2):
        path_suffix = chunks[ii].split('/artwork/', 1)[-1]
        chunks[ii] = '>{0}</'.format(os.path.join(path_prefix, path_suffix).replace('\\', '/'))
        has_changed = True
    # Join into single string and return if something has changed
    if has_changed:
        return ''.join(chunks)
    return None



# -*- coding: utf-8 -*-
"""
Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
from qudi.core.application import Qudi

# parse commandline parameters
parser = argparse.ArgumentParser(prog='python -m qudi.core')
parser.add_argument(
    '-g',
    '--no-gui',
    action='store_true',
    help='Run qudi "headless", i.e. without GUI. User interaction only possible via IPython kernel.'
)
parser.add_argument(
    '-d',
    '--debug',
    action='store_true',
    help='Run qudi in debug mode to log all debug messages. Can affect performance.'
)
parser.add_argument(
    '-c',
    '--config',
    default=None,
    help='Path to the configuration file to use for for this qudi session.'
)
parser.add_argument(
    '-l',
    '--logdir',
    default='',
    help='Absolute path to log directory to use instead of the default one "<user_home>/qudi/log/"'
)
args = parser.parse_args()

app = Qudi(no_gui=args.no_gui, debug=args.debug, log_dir=args.logdir, config_file=args.config)
app.run()



# -*- coding: utf-8 -*-

"""
This file contains a custom QWidget to wrap a collapsible widget with expand/collapse animation.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['CollapsibleWidget']


from PySide2 import QtCore, QtWidgets
from typing import Optional


class CollapsibleWidget(QtWidgets.QWidget):
    """ ToDo: Document
    """

    sigCollapsedChanged = QtCore.Signal(bool)  # True: Collapsed, False: Expanded

    def __init__(self,
                 widget: QtWidgets.QWidget,
                 title: Optional[str] = '',
                 animation_duration: Optional[float] = 0.2,
                 parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent=parent)
        if animation_duration is None:
            animation_duration = 0.2

        self._last_collapsed = True

        self.expand_collapse_button = QtWidgets.QToolButton()
        self.expand_collapse_button.setStyleSheet(
            'QToolButton { border: none; background-color: none; }'
        )
        self.expand_collapse_button.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.expand_collapse_button.setArrowType(QtCore.Qt.RightArrow)
        self.expand_collapse_button.setText(title if title else '')
        self.expand_collapse_button.setCheckable(True)

        widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        widget.setMaximumHeight(0)
        widget.setMinimumHeight(0)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.expand_collapse_button)
        layout.addWidget(widget)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        # Animations to adjust max and min height of this widget and content widget
        self._min_height_animation = QtCore.QPropertyAnimation(self, b"minimumHeight")
        self._max_height_animation = QtCore.QPropertyAnimation(self, b"maximumHeight")
        self._widget_max_height_animation = QtCore.QPropertyAnimation(widget, b"maximumHeight")
        collapsed_height = self.sizeHint().height() - widget.maximumHeight()
        widget_height = widget.sizeHint().height()
        self._min_height_animation.setStartValue(collapsed_height)
        self._min_height_animation.setEndValue(collapsed_height + widget_height)
        self._max_height_animation.setStartValue(collapsed_height)
        self._max_height_animation.setEndValue(collapsed_height + widget_height)
        self._widget_max_height_animation.setStartValue(0)
        self._widget_max_height_animation.setEndValue(widget_height)
        self._animation_group = QtCore.QParallelAnimationGroup()
        self._animation_group.addAnimation(self._min_height_animation)
        self._animation_group.addAnimation(self._max_height_animation)
        self._animation_group.addAnimation(self._widget_max_height_animation)

        self.set_animation_duration(animation_duration)

        self.expand_collapse_button.toggled[bool].connect(self._start_animation)
        self._animation_group.finished.connect(self._animation_finished)

    @property
    def collapsed(self) -> bool:
        if self._min_height_animation.startValue() == self.minimumHeight():
            return True
        if self._min_height_animation.endValue() == self.minimumHeight():
            return False
        return self._last_collapsed

    @property
    def animation_duration(self) -> float:
        return self._min_height_animation.duration() / 1000.

    def set_collapsed(self, collapse: bool) -> None:
        self.expand_collapse_button.setChecked(not collapse)

    def set_animation_duration(self, duration: float) -> None:
        duration_ms = int(round(1000 * duration))
        self._min_height_animation.setDuration(duration_ms)
        self._max_height_animation.setDuration(duration_ms)
        self._widget_max_height_animation.setDuration(duration_ms)

    @QtCore.Slot(bool)
    def _start_animation(self, checked: bool) -> None:
        if checked:
            arrow_type = QtCore.Qt.DownArrow
            direction = QtCore.QAbstractAnimation.Forward
        else:
            arrow_type = QtCore.Qt.RightArrow
            direction = QtCore.QAbstractAnimation.Backward
        self.expand_collapse_button.setArrowType(arrow_type)
        self._animation_group.setDirection(direction)
        self._animation_group.start()

    @QtCore.Slot()
    def _animation_finished(self) -> None:
        collapsed = self.collapsed
        if collapsed != self._last_collapsed:
            self._last_collapsed = collapsed
            self.sigCollapsedChanged.emit(collapsed)



# -*- coding: utf-8 -*-

"""
This file contains modified pyqtgraph ViewBoxes for qudi to track mouse activity and provide
advanced functionality inside data plots.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['MouseTrackingViewBox', 'DataSelectionViewBox', 'RubberbandZoomViewBox',
           'RubberbandZoomSelectionViewBox', 'RubberbandZoomMixin', 'DataSelectionMixin',
           'MouseTrackingMixin', 'SelectionMode']

import warnings
from typing import Optional, Union, Any, Tuple, Sequence, List, Dict
from enum import IntEnum

from PySide2 import QtCore
from pyqtgraph import ViewBox, PlotDataItem, ImageItem, PlotCurveItem, ScatterPlotItem
from pyqtgraph import LinearRegionItem as _LinearRegionItem
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent, MouseDragEvent
from qudi.util.widgets.plotting.marker import Rectangle, InfiniteCrosshairRectangle
from qudi.util.widgets.plotting.marker import InfiniteLine, LinearRegion, InfiniteCrosshair


class SelectionMode(IntEnum):
    Disabled = 0
    X = 1
    Y = 2
    XY = 3


class MouseTrackingMixin:
    """ Extension for pg.ViewBox to tap into mouse move/click/drag events and emit signals.

    x-y-positions emitted will be in real world data coordinates.
    """

    # start_position (x, y), current_position (x, y), MouseDragEvent
    sigMouseDragged = QtCore.Signal(tuple, tuple, object)
    # position (x, y), MouseClickEvent
    sigMouseClicked = QtCore.Signal(tuple, object)

    def __init__(self,
                 allow_tracking_outside_data: Optional[bool] = False,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.allow_tracking_outside_data = bool(allow_tracking_outside_data)

    def mouseClickEvent(self, ev: MouseClickEvent) -> None:
        if self.allow_tracking_outside_data or self.pointer_on_data(ev.scenePos()):
            pos = self.mapToView(ev.pos())
            self.sigMouseClicked.emit((pos.x(), pos.y()), ev)
        if not ev.isAccepted():
            super().mouseClickEvent(ev)

    def mouseDragEvent(self, ev: MouseDragEvent, axis: Optional[int] = None) -> None:
        if self.allow_tracking_outside_data or self.pointer_on_data(ev.buttonDownScenePos()):
            start = self.mapToView(ev.buttonDownPos())
            current = self.mapToView(ev.pos())
            self.sigMouseDragged.emit((start.x(), start.y()), (current.x(), current.y()), ev)
        if not ev.isAccepted():
            super().mouseDragEvent(ev, axis)

    def pointer_on_data(self, scene_pos: QtCore.QPointF) -> bool:
        scene = self.scene()
        if scene is not None:
            for item in scene.items(scene_pos):
                if isinstance(item, (PlotDataItem, ImageItem, PlotCurveItem, ScatterPlotItem)):
                    return True
        return False


class DataSelectionMixin:
    """ Expands MouseTrackingViewBox with data selection functionality.

    You can select:
        - linear region in x or y
        - xy ROIs
        - x, y or xy data markers

    Public methods will generally emit signals if they result in a selection change. The
    corresponding "protected" methods do not emit.
    """

    SelectionMode = SelectionMode

    sigMarkerSelectionChanged = QtCore.Signal(dict)
    sigRegionSelectionChanged = QtCore.Signal(dict)

    def __init__(self,
                 selection_bounds: Optional[Sequence[Tuple[Union[None, float], Union[None, float]]]] = None,
                 selection_pen: Optional[Any] = None,
                 selection_hover_pen: Optional[Any] = None,
                 selection_brush: Optional[Any] = None,
                 selection_hover_brush: Optional[Any] = None,
                 xy_region_selection_crosshair: Optional[bool] = False,
                 xy_region_selection_handles: Optional[bool] = True,
                 xy_region_min_size_percentile: Optional[float] = None,
                 **kwargs
                 ) -> None:
        super().__init__(**kwargs)
        self._selection_bounds = None if selection_bounds is None else list(selection_bounds)
        self._selection_pen = selection_pen
        self._selection_hover_pen = selection_hover_pen
        self._selection_brush = selection_brush
        self._selection_hover_brush = selection_hover_brush
        self._xy_region_selection_crosshair = bool(xy_region_selection_crosshair)
        self._xy_region_selection_handles = bool(xy_region_selection_handles)
        self._region_selection_mode = self.SelectionMode.Disabled
        self._marker_selection_mode = self.SelectionMode.Disabled
        self._selection_mutable = True
        if (xy_region_min_size_percentile is not None) and (xy_region_min_size_percentile > 0):
            self._xy_region_min_size_percentile = xy_region_min_size_percentile
            self.sigRangeChanged.connect(self._update_xy_region_min_size)
        else:
            self._xy_region_min_size_percentile = None

        self.__regions = list()
        self.__markers = list()

    def _update_xy_region_min_size(self, viewbox, new_range, changed) -> None:
        min_size = [
            self._xy_region_min_size_percentile * abs(rang[1] - rang[0]) for rang in new_range
        ]
        for region in self.__regions:
            region.set_min_size(min_size)

    def mouseClickEvent(self, ev: MouseClickEvent) -> None:
        if self.allow_tracking_outside_data or self.pointer_on_data(ev.scenePos()):
            selection_enabled = self._marker_selection_mode != self.SelectionMode.Disabled
            if selection_enabled and (ev.button() == QtCore.Qt.LeftButton) and not ev.double():
                ev.accept()
                pos = self.mapToView(ev.pos())
                self.add_marker_selection((pos.x(), pos.y()))
        return super().mouseClickEvent(ev)

    def mouseDragEvent(self, ev: MouseDragEvent, axis: Optional[int] = None) -> None:
        if not ev.isAccepted():
            selection_enabled = self._region_selection_mode != self.SelectionMode.Disabled
            no_mod = ev.modifiers() == QtCore.Qt.NoModifier
            is_left_button = ev.button() == QtCore.Qt.LeftButton
            data_valid = self.allow_tracking_outside_data or self.pointer_on_data(
                ev.buttonDownScenePos()
            )
            if selection_enabled and no_mod and (axis is None) and is_left_button and data_valid:
                ev.accept()
                start = self.mapToView(ev.buttonDownPos())
                end = self.mapToView(ev.pos())
                span = ((start.x(), end.x()), (start.y(), end.y()))
                if ev.isStart():
                    self._add_region_selection(span)
                else:
                    try:
                        self._move_region_selection(span, index=-1)
                    except IndexError:
                        pass
                if ev.isFinish():
                    self._emit_region_change()
        return super().mouseDragEvent(ev, axis)

    @property
    def region_selection_mode(self) -> SelectionMode:
        return self._region_selection_mode

    def set_region_selection_mode(self, mode: Union[SelectionMode, int]) -> None:
        self._region_selection_mode = self.SelectionMode(mode)

    @property
    def marker_selection_mode(self) -> SelectionMode:
        return self._marker_selection_mode

    def set_marker_selection_mode(self, mode: Union[SelectionMode, int]) -> None:
        self._marker_selection_mode = self.SelectionMode(mode)

    @property
    def selection_mutable(self) -> bool:
        return self._selection_mutable

    def set_selection_mutable(self, mutable: bool) -> None:
        mutable = bool(mutable)
        if mutable is not self._selection_mutable:
            for m in self.__markers:
                m.set_movable(mutable)
            for r in self.__regions:
                r.set_movable(mutable)
                try:
                    r.set_resizable(mutable)
                except AttributeError:
                    pass
            self._selection_mutable = mutable

    @property
    def selection_bounds(self) -> Union[None, List[Union[None, Tuple[float, float]]]]:
        try:
            return self._selection_bounds.copy()
        except AttributeError:
            return self._selection_bounds

    def set_selection_bounds(self,
                             bounds: Union[None, List[Union[None, Tuple[float, float]]]]
                             ) -> None:
        old_bounds = self.selection_bounds
        if bounds != old_bounds:
            self._selection_bounds = bounds
            try:
                self._apply_selection_bounds()
            except:
                self._selection_bounds = old_bounds
                raise

    def add_region_selection(self,
                             span: Tuple[Tuple[float, float], Tuple[float, float]],
                             mode: Optional[Union[SelectionMode, int]] = None,
                             ) -> None:
        self._add_region_selection(span, mode)
        self._emit_region_change()

    def _add_region_selection(self,
                              span: Tuple[Tuple[float, float], Tuple[float, float]],
                              mode: Optional[Union[SelectionMode, int]] = None,
                              ) -> None:
        mode = self._region_selection_mode if mode is None else self.SelectionMode(mode)
        if mode == self.SelectionMode.Disabled:
            return
        if mode == self.SelectionMode.XY:
            x_min, x_max = sorted(span[0])
            y_min, y_max = sorted(span[1])
            x_span = x_max - x_min
            y_span = y_max - y_min
            if self._xy_region_selection_crosshair:
                item_type = InfiniteCrosshairRectangle
            else:
                item_type = Rectangle
            item = item_type(
                viewbox=self,
                position=(x_min + x_span / 2, y_min + y_span / 2),
                size=(x_span, y_span),
                edge_handles=self.selection_mutable and self._xy_region_selection_handles,
                corner_handles=False,
                bounds=self.selection_bounds,
                movable=self.selection_mutable,
                resizable=self.selection_mutable,
                pen=self._selection_pen,
                hover_pen=self._selection_hover_pen
            )
        else:
            if mode == self.SelectionMode.X:
                orientation = QtCore.Qt.Vertical
                bounds = None if self._selection_bounds is None else self._selection_bounds[0]
                values = span[0]
            else:
                orientation = QtCore.Qt.Horizontal
                bounds = None if self._selection_bounds is None else self._selection_bounds[1]
                values = span[1]
            item = LinearRegion(viewbox=self,
                                orientation=orientation,
                                span=values,
                                bounds=bounds,
                                movable=self.selection_mutable,
                                pen=self._selection_pen,
                                hover_pen=self._selection_hover_pen,
                                brush=self._selection_brush,
                                hover_brush=self._selection_hover_brush)
        item.sigAreaChanged.connect(self._emit_region_change)
        item.set_z_value(10)
        self.__regions.append(item)

    def add_marker_selection(self,
                             position: Tuple[float, float],
                             mode: Optional[Union[SelectionMode, int]] = None,
                             ) -> None:
        self._add_marker_selection(position, mode)
        self._emit_marker_change()

    def _add_marker_selection(self,
                              position: Tuple[float, float],
                              mode: Optional[Union[SelectionMode, int]] = None,
                              ) -> None:
        mode = self._marker_selection_mode if mode is None else self.SelectionMode(mode)
        if mode == self.SelectionMode.Disabled:
            return
        elif mode == self.SelectionMode.XY:
            item = InfiniteCrosshair(viewbox=self,
                                     position=position,
                                     bounds=self.selection_bounds,
                                     movable=self.selection_mutable,
                                     pen=self._selection_pen,
                                     hover_pen=self._selection_hover_pen)
        else:
            if mode == self.SelectionMode.X:
                orientation = QtCore.Qt.Vertical
                bounds = None if self._selection_bounds is None else self._selection_bounds[0]
                pos = position[0]
            else:
                orientation = QtCore.Qt.Horizontal
                bounds = None if self._selection_bounds is None else self._selection_bounds[1]
                pos = position[1]
            item = InfiniteLine(viewbox=self,
                                orientation=orientation,
                                position=pos,
                                bounds=bounds,
                                movable=self.selection_mutable,
                                pen=self._selection_pen,
                                hover_pen=self._selection_hover_pen)
        item.sigPositionChanged.connect(self._emit_marker_change)
        item.set_z_value(11)
        self.__markers.append(item)

    def move_region_selection(self,
                              span: Tuple[Tuple[float, float], Tuple[float, float]],
                              index: int
                              ) -> None:
        self._move_region_selection(span, index)
        self._emit_region_change()

    def _move_region_selection(self,
                               span: Tuple[Tuple[float, float], Tuple[float, float]],
                               index: int
                               ) -> None:
        item = self.__regions[index]
        item.blockSignals(True)
        if isinstance(item, LinearRegion):
            if item.orientation == QtCore.Qt.Vertical:
                item.set_area(sorted(span[0]))
            else:
                item.set_area(sorted(span[1]))
        elif isinstance(item, Rectangle):
            x_min, x_max = sorted(span[0])
            y_min, y_max = sorted(span[1])
            size = (x_max - x_min, y_max - y_min)
            position = (x_min + size[0] / 2, y_min + size[1] / 2)
            item.set_area(position, size)
        item.blockSignals(False)

    def move_marker_selection(self,
                              position: Tuple[float, float],
                              index: int
                              ) -> None:
        self._move_marker_selection(position, index)
        self._emit_marker_change()

    def _move_marker_selection(self,
                               position: Tuple[float, float],
                               index: int
                               ) -> None:
        item = self.__markers[index]
        item.blockSignals(True)
        if isinstance(item, InfiniteLine):
            item.set_position(position[0 if item.orientation == QtCore.Qt.Vertical else 1])
        elif isinstance(item, InfiniteCrosshair):
            item.set_position(position)
        item.blockSignals(False)

    def clear_marker_selections(self) -> None:
        if len(self.__markers) != 0:
            try:
                while True:
                    self._delete_marker_selection(-1)
            except IndexError:
                pass
            self._emit_marker_change()

    def delete_marker_selection(self, index: int) -> None:
        self._delete_marker_selection(index)
        self._emit_marker_change()

    def _delete_marker_selection(self, index: int) -> None:
        item = self.__markers.pop(index)
        item.sigPositionChanged.disconnect()
        item.hide()
        item.setParent(None)

    def hide_marker_selections(self) -> None:
        for marker in self.__markers:
            marker.hide()

    def show_marker_selections(self) -> None:
        for marker in self.__markers:
            marker.show()

    def hide_marker_selection(self, index: int) -> None:
        self.__markers[index].hide()

    def show_marker_selection(self, index: int) -> None:
        self.__markers[index].show()

    def clear_region_selections(self) -> None:
        if len(self.__regions) != 0:
            try:
                while True:
                    self._delete_region_selection(-1)
            except IndexError:
                pass
            self._emit_region_change()

    def delete_region_selection(self, index: int) -> None:
        self._delete_region_selection(index)
        self._emit_region_change()

    def _delete_region_selection(self, index: int) -> None:
        item = self.__regions.pop(index)
        item.sigAreaChanged.disconnect()
        item.hide()
        item.setParent(None)

    def hide_region_selections(self) -> None:
        for region in self.__regions:
            region.hide()

    def show_region_selections(self) -> None:
        for region in self.__regions:
            region.show()

    def hide_region_selection(self, index: int) -> None:
        self.__regions[index].hide()

    def show_region_selection(self, index: int) -> None:
        self.__regions[index].show()

    @property
    def marker_selection(self) -> Dict[SelectionMode, List[Union[float, Tuple[float, float]]]]:
        return {
            self.SelectionMode.X: [
                m.position for m in self.__markers if
                isinstance(m, InfiniteLine) and m.orientation == QtCore.Qt.Vertical
            ],
            self.SelectionMode.Y: [
                m.position for m in self.__markers if
                isinstance(m, InfiniteLine) and m.orientation == QtCore.Qt.Horizontal
            ],
            self.SelectionMode.XY: [
                m.position for m in self.__markers if isinstance(m, InfiniteCrosshair)
            ]
        }

    @property
    def region_selection(self) -> Dict[SelectionMode, List[tuple]]:
        return {
            self.SelectionMode.X: [
                r.area for r in self.__regions if
                isinstance(r, LinearRegion) and r.orientation == QtCore.Qt.Vertical
            ],
            self.SelectionMode.Y: [
                r.area for r in self.__regions if
                isinstance(r, LinearRegion) and r.orientation == QtCore.Qt.Horizontal
            ],
            self.SelectionMode.XY: [r.area for r in self.__regions if isinstance(r, Rectangle)]
        }

    def _emit_marker_change(self) -> None:
        self.sigMarkerSelectionChanged.emit(self.marker_selection)

    def _emit_region_change(self) -> None:
        self.sigRegionSelectionChanged.emit(self.region_selection)

    def _apply_selection_bounds(self) -> None:
        if self._selection_bounds is None:
            x_bounds = y_bounds = None
        else:
            x_bounds, y_bounds = self._selection_bounds
        for m in self.__markers:
            if isinstance(m, InfiniteLine):
                m.set_bounds(x_bounds if m.orientation == QtCore.Qt.Vertical else y_bounds)
            elif isinstance(m, InfiniteCrosshair):
                m.set_bounds((x_bounds, y_bounds))
        for r in self.__regions:
            if isinstance(r, LinearRegion):
                r.set_bounds(x_bounds if r.orientation == QtCore.Qt.Vertical else y_bounds)
            elif isinstance(r, Rectangle):
                r.set_bounds((x_bounds, y_bounds))


class RubberbandZoomMixin:
    """
    """
    SelectionMode = SelectionMode

    sigZoomAreaApplied = QtCore.Signal(QtCore.QRectF)

    # FIXME: Introduce general dependency checking during qudi startup to avoid cluttering with
    #  checks like the one below
    try:
        from pyqtgraph import __version__ as __pyqtgraph_version
        if __pyqtgraph_version == '0.12.4':
            raise RuntimeError(
                'You are using an unupported version of pyqtgraph. Please re-install qudi-core '
                'using pip or update pyqtgraph to a version != 0.12.4 manually.'
            )
    except ImportError:
        pass

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._rubberband_zoom_selection_mode = self.SelectionMode.Disabled
        self._x_zoom_region = _LinearRegionItem(orientation='vertical',
                                                brush=kwargs.get('brush', None),
                                                pen=kwargs.get('pen', None),
                                                hoverBrush=kwargs.get('hover_brush', None),
                                                hoverPen=kwargs.get('hover_pen', None),
                                                movable=False)
        self._y_zoom_region = _LinearRegionItem(orientation='horizontal',
                                                brush=kwargs.get('brush', None),
                                                pen=kwargs.get('pen', None),
                                                hoverBrush=kwargs.get('hover_brush', None),
                                                hoverPen=kwargs.get('hover_pen', None),
                                                movable=False)

    @property
    def rubberband_zoom_selection_mode(self) -> SelectionMode:
        return self._rubberband_zoom_selection_mode

    def set_rubberband_zoom_selection_mode(self, mode: SelectionMode) -> None:
        """ Set selection mode for automatic zooming into a rubberband selection when dragging the
        mouse cursor.
        """
        self._rubberband_zoom_selection_mode = self.SelectionMode(mode)

    def mouseDragEvent(self, ev, axis=None):
        """ Additional mouse drag event handling to implement rubber band selection and zooming.
        """
        if not ev.isAccepted():
            no_mod = ev.modifiers() == QtCore.Qt.NoModifier
            is_left_button = ev.button() == QtCore.Qt.LeftButton
            mode = self._rubberband_zoom_selection_mode
            zoom_enabled = mode != self.SelectionMode.Disabled

            if zoom_enabled and is_left_button and no_mod:
                ev.accept()
                super().mouseDragEvent(ev, axis)
                start_pos = self.mapToView(ev.buttonDownPos())
                current_pos = self.mapToView(ev.pos())
                zoom_rect = QtCore.QRectF(start_pos, current_pos)
                if mode == self.SelectionMode.XY:
                    self.updateScaleBox(ev.buttonDownPos(), ev.pos())
                    if ev.isFinish():
                        self.rbScaleBox.hide()
                        self.setRange(rect=zoom_rect, padding=0)
                        self.sigZoomAreaApplied.emit(zoom_rect)
                elif mode == self.SelectionMode.X:
                    self._x_zoom_region.setRegion((start_pos.x(), current_pos.x()))
                    if ev.isStart():
                        self.addItem(self._x_zoom_region)
                    elif ev.isFinish():
                        self.removeItem(self._x_zoom_region)
                        self.setRange(xRange=self._x_zoom_region.getRegion(), padding=0)
                        zoom_rect.setHeight(0)
                        self.sigZoomAreaApplied.emit(zoom_rect)
                elif mode == self.SelectionMode.Y:
                    self._y_zoom_region.setRegion((start_pos.y(), current_pos.y()))
                    if ev.isStart():
                        self.addItem(self._y_zoom_region)
                    elif ev.isFinish():
                        self.removeItem(self._y_zoom_region)
                        self.setRange(yRange=self._y_zoom_region.getRegion(), padding=0)
                        zoom_rect.setWidth(0)
                        self.sigZoomAreaApplied.emit(zoom_rect)
            else:
                super().mouseDragEvent(ev, axis)


class MouseTrackingViewBox(MouseTrackingMixin, ViewBox):
    pass


class DataSelectionViewBox(DataSelectionMixin, MouseTrackingMixin, ViewBox):
    pass


class RubberbandZoomViewBox(RubberbandZoomMixin, MouseTrackingMixin, ViewBox):
    pass


class RubberbandZoomSelectionViewBox(RubberbandZoomMixin, DataSelectionMixin, MouseTrackingMixin, ViewBox):
    pass
